@@ -20,6 +20,58 @@
 #include <math.h>
 #include "gts.h"
 
+#ifdef USE_SURFACE_BTREE
+
+static gint find_closest (GtsTriangle * t, gpointer value, gpointer * data)
+{
+  guint * ns = data[2];
+  guint * n = data[3];
+
+  if (*n >= *ns)
+    return TRUE;
+  else {
+    gdouble * dmin = data[0];
+    gpointer * closest = data[1];
+    GtsPoint * p = data[4];
+
+    if (gts_triangle_orientation (t) > 0.) {
+      gdouble d = gts_point_distance2 (p, GTS_POINT (GTS_SEGMENT (t->e1)->v1));
+      
+      if (d < *dmin) {
+	*dmin = d;
+	*closest = t;
+      }
+      (*n)++;
+    }
+  }
+  return FALSE;
+}
+
+/* select the face closest to @p among n^1/3 randomly picked faces
+ *  of @surface */
+static GtsFace * closest_face (GtsSurface * s, GtsPoint * p)
+{
+  guint n = 0, nt, ns;
+  gdouble dmin = G_MAXDOUBLE;
+  GtsFace * closest = NULL;
+  gpointer data[5];
+
+  nt = gts_surface_face_number (s);
+  if (!nt)
+    return NULL;
+  ns = exp (log ((gdouble) nt)/3.);
+
+  data[0] = &dmin;
+  data[1] = &closest;
+  data[2] = &ns;
+  data[3] = &n;
+  data[4] = p;
+  g_tree_traverse (s->faces, (GTraverseFunc) find_closest, G_IN_ORDER, data);
+  return closest;
+}
+
+#else /* not USE_SURFACE_BTREE */
+
 /* Due to an unkown reason g_hash_table_foreach does not allow to stop 
  * the loop, hence the redefinition. I hope they don't change
  * the GHashTable, GHashNode structures ... */
@@ -32,6 +84,7 @@ struct _GHashNode
   GHashNode *next;
 };
 
+#if GLIB_CHECK_VERSION(1,2,8)
 struct _GHashTable
 {
   gint size;
@@ -41,13 +94,27 @@ struct _GHashTable
   GHashFunc hash_func;
   GCompareFunc key_compare_func;
 };
+#else  /* not 1.2.8 */
+#  if GLIB_CHECK_VERSION(1,3,0)
+struct _GHashTable
+{
+  gint size;
+  gint nnodes;
+  GHashNode **nodes;
+  GHashFunc hash_func;
+  GEqualFunc key_equal_func;
+};
+#  else  /* not 1.3.0 */
+#    error "Hack needs glib version 1.2.8 or 1.3.0"
+#  endif /* not 1.3.0 */
+#endif /* not 1.2.8 */
 
 /* select the face closest to @p among n^1/3 randomly picked faces
  *  of @surface */
 static GtsFace * closest_face (GtsSurface * s, GtsPoint * p)
 {
   guint i, n, nt, ns;
-  gdouble dmin = G_MAXDOUBLE, d;
+  gdouble dmin = G_MAXDOUBLE;
   GtsFace * closest = NULL;
   GHashNode * node;
   GHashTable * hash_table = s->faces;
@@ -58,19 +125,25 @@ static GtsFace * closest_face (GtsSurface * s, GtsPoint * p)
   ns = exp(log((gdouble) nt)/3.);
 
   for (i = 0, n = 0; i < hash_table->size && n < ns; i++)
-    for (node = hash_table->nodes[i]; node && n < ns; node = node->next, n++) {
+    for (node = hash_table->nodes[i]; node && n < ns; node = node->next) {
       GtsFace * f = node->key;
 
-      d = gts_point_distance2 (p, 
-		   GTS_POINT (GTS_SEGMENT (GTS_TRIANGLE (f)->e1)->v1));
-      if (d < dmin) {
-	dmin = d;
-	closest = f;
+      if (gts_triangle_orientation (GTS_TRIANGLE (f)) > 0.) {
+	gdouble d = gts_point_distance2 (p, 
+		      GTS_POINT (GTS_SEGMENT (GTS_TRIANGLE (f)->e1)->v1));
+
+	if (d < dmin) {
+	  dmin = d;
+	  closest = f;
+	}
+	n++;
       }
     }
   return closest;
 }
 
+#endif /* not USE_SURFACE_BTREE */
+
 /* returns the face belonging to @surface and neighbor of @f via @e */
 static GtsFace * neighbor (GtsFace * f,
 			   GtsEdge * e,
@@ -273,6 +346,9 @@ GtsFace * gts_point_locate (GtsPoint * p,
 
   if (guess == NULL)
     guess = closest_face (surface, p);
+  else
+    g_return_val_if_fail (gts_triangle_orientation (GTS_TRIANGLE (guess)) > 0.,
+			  NULL);
   if (guess == NULL)
     return NULL;
 
@@ -358,20 +434,24 @@ static void swap_if_in_circle (GtsFace * f1,
 			   GTS_POINT (v2), GTS_POINT (v3)) > 0.0) {
     GtsEdge * en;
     GtsSegment * sn = gts_vertices_are_connected (v3, v4);
+    GtsFace * f3, * f4;
 
     if (!GTS_IS_EDGE (sn))
       en = gts_edge_new (surface->edge_class, v3, v4);
     else
       en = GTS_EDGE (sn);
 
-    f1 = gts_face_new (surface->face_class, en, e5, e2);
-    gts_surface_add_face (surface, f1);
-    f2 = gts_face_new (surface->face_class, en, e3, e4);
-    gts_surface_add_face (surface, f2);
-    gts_object_destroy (GTS_OBJECT (e1));
+    f3 = gts_face_new (surface->face_class, en, e5, e2);
+    gts_object_attributes (GTS_OBJECT (f3), GTS_OBJECT (f1));
+    f4 = gts_face_new (surface->face_class, en, e3, e4);
+    gts_object_attributes (GTS_OBJECT (f4), GTS_OBJECT (f2));
+    gts_surface_remove_face (surface, f1);
+    gts_surface_remove_face (surface, f2);
+    gts_surface_add_face (surface, f3);
+    gts_surface_add_face (surface, f4);
 
-    swap_if_in_circle (f1, v4, v2, v3, e5, e2, en, surface);
-    swap_if_in_circle (f2, v1, v4, v3, e4, en, e3, surface);
+    swap_if_in_circle (f3, v4, v2, v3, e5, e2, en, surface);
+    swap_if_in_circle (f4, v1, v4, v3, e4, en, e3, surface);
   }
 }
 
@@ -434,8 +514,11 @@ GtsVertex * gts_delaunay_add_vertex_to_face (GtsSurface * surface,
 
   /* cf. figure misc/swap.fig */
   f1 = gts_face_new (surface->face_class, e4, e1, e5);
+  gts_object_attributes (GTS_OBJECT (f1), GTS_OBJECT (f));
   f2 = gts_face_new (surface->face_class, e5, e2, e6);
+  gts_object_attributes (GTS_OBJECT (f2), GTS_OBJECT (f));
   f3 = gts_face_new (surface->face_class, e6, e3, e4);
+  gts_object_attributes (GTS_OBJECT (f3), GTS_OBJECT (f));
 
   gts_surface_remove_face (surface, f);
   gts_surface_add_face (surface, f1);
@@ -510,13 +593,16 @@ static gboolean polygon_in_circle (GSList * poly,
   return FALSE;
 }
 
-static void triangulate_polygon (GSList * poly, GtsSurface * surface)
+static void triangulate_polygon (GSList * poly, 
+				 GtsSurface * surface,
+				 GtsFace * ref)
 {
   GSList * i, * poly1, * poly2;
   GtsVertex * v1, * v2, * v3 = NULL;
   gboolean found = FALSE;
   GtsSegment * s, * s1, * s2;
   GtsEdge * e1, * e2;
+  GtsFace * f;
 
   if (poly == NULL || poly->next == NULL) {
     g_slist_free (poly);
@@ -573,8 +659,9 @@ static void triangulate_polygon (GSList * poly, GtsSurface * surface)
     e2 = gts_edge_new (surface->edge_class, v3, v1);
   else
     e2 = GTS_EDGE (s2);
-  gts_surface_add_face (surface,
-	    gts_face_new (surface->face_class, GTS_EDGE (s), e1, e2));
+  f = gts_face_new (surface->face_class, GTS_EDGE (s), e1, e2);
+  gts_object_attributes (GTS_OBJECT (f), GTS_OBJECT (ref));
+  gts_surface_add_face (surface, f);
 
   poly1 = poly->next;
   g_slist_free_1 (poly);
@@ -587,8 +674,8 @@ static void triangulate_polygon (GSList * poly, GtsSurface * surface)
   else
     i->next = NULL;
 
- triangulate_polygon (poly1, surface);
- triangulate_polygon (poly2, surface);
+ triangulate_polygon (poly1, surface, ref);
+ triangulate_polygon (poly2, surface, ref);
 }
 
 /**
@@ -596,24 +683,43 @@ static void triangulate_polygon (GSList * poly, GtsSurface * surface)
  * @surface: a #GtsSurface.
  * @v: a #GtsVertex.
  *
- * Removes @v from the Delaunay triangulation defined by @surface,
- * restores the Delaunay property and destroys @v. Vertex @v must not
- * be used by any constrained edge otherwise the triangulation is not
- * guaranteed to be Delaunay.  
+ * Removes @v from the Delaunay triangulation defined by @surface and
+ * restores the Delaunay property. Vertex @v must not be used by any
+ * constrained edge otherwise the triangulation is not guaranteed to
+ * be Delaunay.  
  */
 void gts_delaunay_remove_vertex (GtsSurface * surface, GtsVertex * v)
 {
+  GSList * triangles, * i;
+  GtsFace * ref = NULL;
+
   g_return_if_fail (surface != NULL);
   g_return_if_fail (v != NULL);
 
-  triangulate_polygon (gts_vertex_fan_oriented (v, surface), surface);
-  gts_object_destroy (GTS_OBJECT (v));
+  i = triangles = gts_vertex_triangles (v, NULL);
+  while (i && !ref) {
+    if (GTS_IS_FACE (i->data) &&
+	gts_face_has_parent_surface (i->data, surface))
+      ref = i->data;
+    i = i->next;
+  }
+  if (!ref) {
+    g_slist_free (triangles);
+    g_return_if_fail (ref);
+  }
+  triangulate_polygon (gts_vertex_fan_oriented (v, surface), surface, ref);
+  while (i) {
+    if (GTS_IS_FACE (i->data) &&
+	gts_face_has_parent_surface (i->data, surface))
+      gts_surface_remove_face (surface, i->data);
+    i = i->next;
+  }
+  g_slist_free (triangles);
 }
 
 #define NEXT_CUT(edge, edge1, list) { next = neighbor (f, edge, surface);\
-                                      if (constraint)\
-					destroy_triangles (e);\
-				      else\
+                                      remove_triangles (e, surface);\
+                                      if (!constraint && !e->triangles)\
 				        gts_object_destroy (GTS_OBJECT (e));\
                                       g_assert (next);\
 				      *list = g_slist_prepend (*list, edge1);\
@@ -622,16 +728,17 @@ void gts_delaunay_remove_vertex (GtsSurface * surface, GtsVertex * v)
 					       next, surface, left, right));\
                                     }
 
-static void destroy_triangles (GtsEdge * e)
+static void remove_triangles (GtsEdge * e, GtsSurface * s)
 {
   GSList * i = e->triangles;
+
   while (i) {
     GSList * next = i->next;
-    gts_object_destroy (i->data);
+
+    if (GTS_IS_FACE (i->data) && gts_face_has_parent_surface (i->data, s))
+      gts_surface_remove_face (s, i->data);
     i = next;
   }
-  g_slist_free (e->triangles);
-  e->triangles = NULL;
 }
 
 static GSList * 
@@ -661,9 +768,8 @@ remove_intersected_edge (GtsSegment * s,
   if (o1 >= 0.) {
     if (o2 >= 0.) { /* @s->v2 is inside (or on the edge) of @f */
       g_assert (o1 == 0. && o2 == 0.);
-      if (constraint)
-	destroy_triangles (e);
-      else
+      remove_triangles (e, surface);
+      if (!constraint && !e->triangles)
 	gts_object_destroy (GTS_OBJECT (e));
       *left = g_slist_prepend (*left, e2);
       *right = g_slist_prepend (*right, e1);
@@ -689,7 +795,8 @@ remove_intersected_vertex (GtsSegment * s,
 			   GtsVertex * v,
 			   GtsSurface * surface,
 			   GSList ** left,
-			   GSList ** right)
+			   GSList ** right,
+			   GtsFace ** ref)
 {
   GSList * triangles = gts_vertex_triangles (v, NULL);
   GSList * i;
@@ -723,7 +830,8 @@ remove_intersected_vertex (GtsSegment * s,
 	GtsEdge * e = gts_triangle_edge_opposite (t, v);
 	GtsEdge * e1, * e2;
 	GtsFace * next = neighbor (GTS_FACE (t), e, surface);
-	
+
+	*ref = GTS_FACE (t);
 	gts_triangle_vertices_edges (t, e, &v2, &v3, &v, &e, &e2, &e1);
 
 	g_slist_free (triangles);
@@ -731,7 +839,7 @@ remove_intersected_vertex (GtsSegment * s,
 	if (o3 >= 0.) /* @s->v2 is inside (or on the edge) of t */
 	  return NULL;
 
-	gts_object_destroy (GTS_OBJECT (t));
+	gts_surface_remove_face (surface, GTS_FACE (t));
 	*left = g_slist_prepend (*left, e2);
 	*right = g_slist_prepend (*right, e1);
 
@@ -755,7 +863,7 @@ remove_intersected_vertex (GtsSegment * s,
  * @surface.
  *
  * Returns: a list of #GtsConstraint conflicting (i.e. intersecting) with @c 
- * which were removed from @surface.
+ * which were removed from @surface (%NULL if there was none).
  */
 GSList * gts_delaunay_add_constraint (GtsSurface * surface,
 				      GtsConstraint * c)
@@ -763,6 +871,7 @@ GSList * gts_delaunay_add_constraint (GtsSurface * surface,
   GSList * constraints;
   GtsVertex * v1, * v2;
   GSList * left = NULL, * right = NULL;
+  GtsFace * ref = NULL;
 
   g_return_val_if_fail (surface != NULL, NULL);
   g_return_val_if_fail (c != NULL, NULL);
@@ -773,11 +882,13 @@ GSList * gts_delaunay_add_constraint (GtsSurface * surface,
   
   gts_allow_floating_edges = TRUE;
   constraints = remove_intersected_vertex (GTS_SEGMENT (c), v1, surface,
-					   &left, &right);
+					   &left, &right, &ref);
   gts_allow_floating_edges = FALSE;
 #if 1
-  triangulate_polygon (g_slist_prepend (g_slist_reverse (right), c), surface);
-  triangulate_polygon (g_slist_prepend (left, c), surface);
+  triangulate_polygon (g_slist_prepend (g_slist_reverse (right), c), 
+		       surface, ref);
+  triangulate_polygon (g_slist_prepend (left, c), 
+		       surface, ref);
 #else
   right = g_slist_prepend (g_slist_reverse (right), c);
   left = g_slist_prepend (left, c);
@@ -897,6 +1008,7 @@ void gts_delaunay_remove_hull (GtsSurface * surface)
     g_slist_free_1 (i);
     if (!GTS_IS_CONSTRAINT (e)) {
       GtsTriangle * t = GTS_TRIANGLE (gts_edge_is_boundary (e, surface));
+
       if (t != NULL) {
 	if (t->e1 != e && !GTS_IS_CONSTRAINT (t->e1) &&
 	    !gts_edge_is_boundary (t->e1, surface))
@@ -907,8 +1019,10 @@ void gts_delaunay_remove_hull (GtsSurface * surface)
 	if (t->e3 != e && !GTS_IS_CONSTRAINT (t->e3) &&
 	    !gts_edge_is_boundary (t->e3, surface))
 	  boundary = g_slist_prepend (boundary, t->e3);
+	gts_surface_remove_face (surface, GTS_FACE (t));
       }
-      gts_object_destroy (GTS_OBJECT (e));
+      if (!e->triangles)
+	gts_object_destroy (GTS_OBJECT (e));
     }
   }
   gts_allow_floating_edges = FALSE;
