@@ -28,7 +28,9 @@ gboolean gts_allow_floating_gnodes = FALSE;
 static void gnode_remove_container (GtsContainee * i, GtsContainer * c)
 {
   (* GTS_CONTAINEE_CLASS (GTS_OBJECT_CLASS (gts_gnode_class ())->parent_class)->remove_container) (i, c);
-  if (GTS_SLIST_CONTAINEE (i)->containers == NULL && !gts_allow_floating_gnodes)
+  if (GTS_SLIST_CONTAINEE (i)->containers == NULL && 
+      !gts_allow_floating_gnodes &&
+      !GTS_OBJECT_DESTROYED(GTS_OBJECT (i)))
     gts_object_destroy (GTS_OBJECT (i));
 }
 
@@ -340,8 +342,83 @@ GtsWGNode * gts_wgnode_new (GtsWGNodeClass * klass,
   return n;
 }
 
+/* GtsPNode */
+
+static void pnode_write (GtsGNode * n, FILE * fp)
+{
+  if (GTS_IS_NVERTEX (GTS_PNODE (n)->data))
+    fprintf (fp, "label=\"%p:%s\",", 
+	     GTS_PNODE (n)->data,
+	     GTS_NVERTEX (GTS_PNODE (n)->data)->name);
+  else
+    fprintf (fp, "label=\"%p\",", GTS_PNODE (n)->data);
+}
+
+static void pnode_class_init (GtsPNodeClass * klass)
+{
+  GTS_GNODE_CLASS (klass)->write = pnode_write;
+}
+
+static void pnode_init (GtsPNode * pn)
+{
+  pn->data = NULL;
+}
+
+/**
+ * gts_pnode_class:
+ *
+ * Returns: the #GtsPNodeClass.
+ */
+GtsPNodeClass * gts_pnode_class (void)
+{
+  static GtsPNodeClass * klass = NULL;
+
+  if (klass == NULL) {
+    GtsObjectClassInfo pnode_info = {
+      "GtsPNode",
+      sizeof (GtsPNode),
+      sizeof (GtsPNodeClass),
+      (GtsObjectClassInitFunc) pnode_class_init,
+      (GtsObjectInitFunc) pnode_init,
+      (GtsArgSetFunc) NULL,
+      (GtsArgGetFunc) NULL
+    };
+    klass = gts_object_class_new (GTS_OBJECT_CLASS (gts_gnode_class ()),
+				  &pnode_info);
+  }
+
+  return klass;
+}
+
+/**
+ * gts_pnode_new:
+ * @klass: a #GtsPNodeClass.
+ * @data: user data.
+ *
+ * Returns: a new #GtsPNode associated with @data.
+ */
+GtsPNode * gts_pnode_new (GtsPNodeClass * klass, gpointer data)
+{
+  GtsPNode * pn;
+
+  pn = GTS_PNODE (gts_object_new (GTS_OBJECT_CLASS (klass)));
+  pn->data = data;
+
+  return pn;
+}
+
 /* GtsFNode */
 
+static void fnode_write (GtsGNode * n, FILE * fp)
+{
+  fprintf (fp, "label=\"%p\",", GTS_FNODE (n)->f);
+}
+
+static void fnode_class_init (GtsGNodeClass * klass)
+{
+  klass->write = fnode_write;
+}
+
 static void fnode_init (GtsFNode * fn)
 {
   fn->f = NULL;
@@ -361,7 +438,7 @@ GtsFNodeClass * gts_fnode_class (void)
       "GtsFNode",
       sizeof (GtsFNode),
       sizeof (GtsFNodeClass),
-      (GtsObjectClassInitFunc) NULL,
+      (GtsObjectClassInitFunc) fnode_class_init,
       (GtsObjectInitFunc) fnode_init,
       (GtsArgSetFunc) NULL,
       (GtsArgGetFunc) NULL
@@ -415,10 +492,10 @@ static void gedge_remove_container (GtsContainee * i, GtsContainer * c)
   ge->n1 = ge->n2 = NULL;
   if (n1 != NULL && n2 != NULL) {
     if (GTS_CONTAINER (n1) == c) {
-      if (n2) gts_container_remove (GTS_CONTAINER (n2), i);
+      if (n2 && n2 != n1) gts_container_remove (GTS_CONTAINER (n2), i);
     }
     else if (GTS_CONTAINER (n2) == c) {
-      if (n1) gts_container_remove (GTS_CONTAINER (n1), i);
+      if (n1 && n1 != n2) gts_container_remove (GTS_CONTAINER (n1), i);
     }
     else
       g_assert_not_reached ();
@@ -438,6 +515,7 @@ static gboolean gedge_is_contained (GtsContainee * i, GtsContainer * c)
 
 static void gedge_class_init (GtsGEdgeClass * klass)
 {
+  klass->link = NULL;
   klass->weight = NULL;
 
   GTS_CONTAINEE_CLASS (klass)->remove_container = gedge_remove_container;
@@ -491,13 +569,16 @@ GtsGEdge * gts_gedge_new (GtsGEdgeClass * klass, GtsGNode * n1, GtsGNode * n2)
 
   g_return_val_if_fail (n1 != NULL, NULL);
   g_return_val_if_fail (n2 != NULL, NULL);
-  g_return_val_if_fail (n1 != n2, NULL);
 
   object = GTS_GEDGE (gts_object_new (GTS_OBJECT_CLASS (klass)));
   object->n1 = n1;
   gts_container_add (GTS_CONTAINER (n1), GTS_CONTAINEE (object));
   object->n2 = n2;
-  gts_container_add (GTS_CONTAINER (n2), GTS_CONTAINEE (object));
+  if (n1 != n2)
+    gts_container_add (GTS_CONTAINER (n2), GTS_CONTAINEE (object));
+
+  if (klass->link)
+    object = (* klass->link) (object, n1, n2);
 
   return object;
 }
@@ -518,6 +599,86 @@ gfloat gts_gedge_weight (GtsGEdge * e)
   return 1.;
 }
 
+/* GtsPGEdge */
+
+static void pgedge_write (GtsGEdge * ge, FILE * fp)
+{
+  if (GTS_IS_EDGE (GTS_PGEDGE (ge)->data)) {
+    GtsEdge * e = GTS_PGEDGE (ge)->data;
+    guint n = g_slist_length (e->triangles);
+
+    fprintf (fp, "label=\"%p:%s:%d\",color=%s", e,
+	     GTS_IS_NEDGE (e) ? GTS_NEDGE (e)->name : "",
+	     n,
+	     n == 0 ? "black" : 
+             n == 1 ? "blue" :
+	     n == 2 ? "green" :
+	     n == 3 ? "violet" :
+	     n == 4 ? "red" : 
+	     "pink");
+  }
+  else
+    fprintf (fp, "label=\"%p\",", GTS_PGEDGE (ge)->data);
+}
+
+static void pgedge_class_init (GtsPGEdgeClass * klass)
+{
+  GTS_GEDGE_CLASS (klass)->write = pgedge_write;
+}
+
+static void pgedge_init (GtsPGEdge * e)
+{
+  e->data = NULL;
+}
+
+/**
+ * gts_pgedge_class:
+ * 
+ * Returns: the #GtsPGEdgeClass.
+ */
+GtsPGEdgeClass * gts_pgedge_class (void)
+{
+  static GtsPGEdgeClass * klass = NULL;
+
+  if (klass == NULL) {
+    GtsObjectClassInfo pgedge_info = {
+      "GtsPGEdge",
+      sizeof (GtsPGEdge),
+      sizeof (GtsPGEdgeClass),
+      (GtsObjectClassInitFunc) pgedge_class_init,
+      (GtsObjectInitFunc) pgedge_init,
+      (GtsArgSetFunc) NULL,
+      (GtsArgGetFunc) NULL
+    };
+    klass = gts_object_class_new (GTS_OBJECT_CLASS (gts_gedge_class ()),
+				  &pgedge_info);
+  }
+
+  return klass;
+}
+
+/**
+ * gts_pgedge_new:
+ * @klass: a #GtsPGEdgeClass.
+ * @n1: a #GtsGNode.
+ * @n2: another #GtsGNode.
+ * @data: user data.
+ *
+ * Returns: a new #GtsPGEdge associated with @data linking @n1 and @n2.
+ */ 
+GtsPGEdge * gts_pgedge_new (GtsPGEdgeClass * klass,
+			    GtsGNode * g1,
+			    GtsGNode * g2,
+			    gpointer data)
+{
+  GtsPGEdge * we;
+
+  we = GTS_PGEDGE (gts_gedge_new (GTS_GEDGE_CLASS (klass), g1, g2));
+  we->data = data;
+
+  return we;
+}
+
 /* GtsWGEdge */
 
 static gfloat wgedge_weight (GtsGEdge * e)
@@ -591,9 +752,43 @@ static void graph_init (GtsGraph * g)
   g->edge_class = gts_gedge_class ();
 }
 
+static void graph_write (GtsObject * object, FILE * fp)
+{
+  GtsGraph * graph = GTS_GRAPH (object);
+
+  fprintf (fp, " %s %s %s",
+	   object->klass->info.name,
+	   GTS_OBJECT_CLASS (graph->node_class)->info.name,
+	   GTS_OBJECT_CLASS (graph->edge_class)->info.name);
+}
+
+static void graph_read (GtsObject ** object, GtsFile * f)
+{
+  if (f->type != GTS_STRING) {
+    gts_file_error (f, "expecting a string (GtsGraphClass)");
+    return;
+  }
+  gts_file_next_token (f);
+
+  if (f->type != GTS_STRING) {
+    gts_file_error (f, "expecting a string (GtsGNodeClass)");
+    return;
+  }
+  gts_file_next_token (f);
+
+  if (f->type != GTS_STRING) {
+    gts_file_error (f, "expecting a string (GtsGEdgeClass)");
+    return;
+  }
+  gts_file_next_token (f);
+}
+
 static void graph_class_init (GtsGraphClass * klass)
 {
   klass->weight = NULL;
+
+  GTS_OBJECT_CLASS (klass)->write = graph_write;
+  GTS_OBJECT_CLASS (klass)->read = graph_read;
 }
 
 /**
@@ -641,6 +836,8 @@ GtsGraph * gts_graph_new (GtsGraphClass * klass,
   g_return_val_if_fail (edge_class != NULL, NULL);
 
   g = GTS_GRAPH (gts_object_new (GTS_OBJECT_CLASS (klass)));
+  g->node_class = node_class;
+  g->edge_class = edge_class;
 
   return g;
 }
@@ -795,7 +992,7 @@ void gts_graph_traverse_destroy (GtsGraphTraverse * t)
 
 static void edge_foreach_node (GtsGNode * n, gpointer * info)
 {
-  GtsFunc func = info[0];
+  GtsFunc func = (GtsFunc) info[0];
   gpointer data = info[1];
   GHashTable * hash = info[2];
   GSList * i = GTS_SLIST_CONTAINER (n)->items;
@@ -998,9 +1195,9 @@ gfloat gts_graph_edges_cut_weight (GtsGraph * g)
 }
 
 /**
- * gts_graph_read:
+ * gts_graph_read_jostle:
  * @g: a #GtsGraph.
- * @fp: a file pointer.
+ * @fp: a #GtsFile.
  *
  * Adds to @g the nodes and edges defined in the file pointed to by
  * @fp. This file must use the Jostle "graph" ASCII format.  
@@ -1008,68 +1205,307 @@ gfloat gts_graph_edges_cut_weight (GtsGraph * g)
  * line number at which they appear in @fp.
  *
  * Returns: 0 if the lecture was successful, the line number at which
- * an error occured otherwise.  
+ * an error occured otherwise (in which case the @error field of @fp
+ * is set).  
  */
-guint gts_graph_read (GtsGraph * g, FILE * fp)
+guint gts_graph_read_jostle (GtsGraph * g, GtsFile * fp)
 {
-  guint nn, ne, line_number = 1, n;
-  GString * buf;
-  gint delim;
+  guint nn, ne, n;
   GtsGNode ** nodes;
-  gboolean syntax_error = FALSE;
 
   g_return_val_if_fail (g != NULL, 1);
   g_return_val_if_fail (fp != NULL, 1);
 
-  buf = g_string_new ("");
+  if (fp->type != GTS_INT) {
+    gts_file_error (fp, "expecting an integer (number of nodes)");
+    return fp->line;
+  }
+  nn = atoi (fp->token->str);  
+  gts_file_next_token (fp);
 
-  delim = gts_get_token (fp, buf, " \t\n", "");
-  if (delim == EOF || delim == '\n') {
-    g_string_free (buf, TRUE);
-    return 1;
+  if (fp->type != GTS_INT) {
+    gts_file_error (fp, "expecting an integer (number of edges)");
+    return fp->line;
   }
-  nn = atoi (buf->str);
-  delim = gts_get_token (fp, buf, " \t\n", "");
-  if (delim == EOF) {
-    g_string_free (buf, TRUE);
-    return 1;
-  }
-  ne = atoi (buf->str);
-  if (delim != '\n')
-    gts_get_newline (fp);
+  ne = atoi (fp->token->str);
+
+  gts_file_first_token_after (fp, '\n');
   nodes = g_malloc (sizeof (GtsGNode *)*(nn + 1));
 
   n = 0;
-  while (n < nn && !syntax_error) {
-    GtsNGNode * node = gts_ngnode_new (gts_ngnode_class (), line_number);
+  while (n < nn && fp->type != GTS_ERROR) {
+    GtsNGNode * node = gts_ngnode_new (gts_ngnode_class (), fp->line);
     
     nodes[n++] = GTS_GNODE (node);
     gts_container_add (GTS_CONTAINER (g), GTS_CONTAINEE (node));
-    line_number++;
     do {
-      delim = gts_get_token (fp, buf, " \t\n", "");
-      if (delim == EOF)
-	syntax_error = TRUE;
+      if (fp->type != GTS_INT)
+	gts_file_error (fp, "expecting an integer (node index)");
       else {
-	guint in = atoi (buf->str);
-	if (in > nn || in == n || in == 0)
-	  syntax_error = TRUE;
+	guint in = atoi (fp->token->str);
+	
+	if (in == 0 || in > nn)
+	  gts_file_error (fp, "node index `%d' is out of range `[1,%d]'",
+			  in, nn);
+	else if (in == n)
+	  gts_file_error (fp, "node index `%d' references itself", in);
 	else if (in < n) {
 	  gts_gedge_new (g->edge_class, GTS_GNODE (node), nodes[in - 1]);
 	  ne--;
+	  gts_file_next_token (fp);
 	}
       }
-    } while (delim != EOF && delim != '\n' && !syntax_error);
+    } while (fp->type != GTS_ERROR && fp->type != '\n');
   }
   g_free (nodes);
-  g_string_free (buf, TRUE);
 
-  if (n != nn || ne > 0 || syntax_error)
-    return line_number;  
+  if (fp->type != GTS_ERROR) {
+    if (n != nn)
+      gts_file_error (fp, "only `%d' nodes read out of `%d'",
+		      n, nn);
+    else if (ne > 0)
+      gts_file_error (fp, "`%d' unallocated edges remaining",
+		      ne);
+  }
 
+  if (fp->type == GTS_ERROR)
+    return fp->line;
   return 0;
 }
 
+static void count_edges (GtsGEdge * e, guint * nedge)
+{
+  (*nedge)++;
+}
+
+static void write_node (GtsObject * node, gpointer * data)
+{
+  FILE * fp = data[0];
+  guint * nnode = data[1];
+
+  node->reserved = GUINT_TO_POINTER ((*nnode)++);
+  if (node->klass->write)
+    (* node->klass->write) (node, fp);
+  fputc ('\n', fp);
+}
+
+static void write_edge (GtsGEdge * edge, FILE * fp)
+{
+  fprintf (fp, "%u %u", 
+	   GPOINTER_TO_UINT (GTS_OBJECT (edge->n1)->reserved),
+	   GPOINTER_TO_UINT (GTS_OBJECT (edge->n2)->reserved));
+  if (GTS_OBJECT (edge)->klass->write)
+    (* GTS_OBJECT (edge)->klass->write) (GTS_OBJECT (edge), fp);
+  fputc ('\n', fp);
+}
+
+/**
+ * gts_graph_write:
+ * @g: a #GtsGraph.
+ * @fp: a file pointer.
+ *
+ * Writes in the file @fp an ASCII representation of @g. The file
+ * format is as follows. 
+ *
+ * All the lines beginning with #GTS_COMMENTS are ignored. The first line
+ * contains two unsigned integers separated by spaces. The first
+ * integer is the number of nodes, nn, the second is the number of
+ * edges, ne.
+ *
+ * Follows nn lines containing node description.
+ * Follows ne lines containing the two indices (starting
+ * from one) of the nodes of each edge.
+ *
+ * The format described above is the least common denominator to all
+ * GTS files.  Consistent with an object-oriented approach, the GTS
+ * file format is extensible. Each of the lines of the file can be
+ * extended with user-specific attributes accessible through the
+ * read() and write() virtual methods of each of the objects written
+ * (graph, nodes or edges). When read with different object classes,
+ * these extra attributes are just ignored.  
+ */
+void gts_graph_write (GtsGraph * g, FILE * fp)
+{
+  guint nnode = 1, nedge = 0;
+  gpointer data[2];
+
+  g_return_if_fail (g != NULL);
+  g_return_if_fail (fp != NULL);
+
+  gts_graph_foreach_edge (g, (GtsFunc) count_edges, &nedge);
+  fprintf (fp, "%u %u", gts_container_size (GTS_CONTAINER (g)), nedge);
+  if (GTS_OBJECT (g)->klass->write)
+    (* GTS_OBJECT (g)->klass->write) (GTS_OBJECT (g), fp);
+  fputc ('\n', fp);
+  data[0] = fp;
+  data[1] = &nnode;
+  gts_container_foreach (GTS_CONTAINER (g), (GtsFunc) write_node, data);
+  gts_graph_foreach_edge (g, (GtsFunc) write_edge, fp);
+  gts_container_foreach (GTS_CONTAINER (g), 
+			 (GtsFunc) gts_object_reset_reserved, NULL);
+}
+
+/**
+ * gts_graph_read:
+ * @g: a #GtsGraph.
+ * @fp: a #GtsFile.
+ *
+ * Adds to @g the data read from @fp. The format of the file pointed
+ * to by @fp is as described in gts_graph_write().
+ *
+ * Returns: 0 if successful or the line number at which the parsing
+ * stopped in case of error (in which case the @error field of @fp is set).
+ */
+guint gts_graph_read (GtsGraph * g, GtsFile * fp)
+{
+  GtsGNode ** nodes;
+  guint nn, ne, n;
+
+  g_return_val_if_fail (g != NULL, 1);
+  g_return_val_if_fail (fp != NULL, 1);
+
+  if (fp->type != GTS_INT) {
+    gts_file_error (fp, "expecting an integer (number of nodes)");
+    return fp->line;
+  }
+  nn = atoi (fp->token->str);
+  gts_file_next_token (fp);
+
+  if (fp->type != GTS_INT) {
+    gts_file_error (fp, "expecting an integer (number of edges)");
+    return fp->line;
+  }
+  ne = atoi (fp->token->str);
+
+  gts_file_next_token (fp);
+  if (fp->type != '\n')
+    if (GTS_OBJECT (g)->klass->read)
+      (* GTS_OBJECT (g)->klass->read) ((GtsObject **) &g, fp);
+  gts_file_first_token_after (fp, '\n');
+  if (nn <= 0)
+    return 0;
+
+  nodes = g_malloc ((nn + 1)*sizeof (GtsGNode *));
+
+  n = 0;
+  while (n < nn && fp->type != GTS_ERROR) {
+    GtsObject * new_node = 
+      gts_object_new (GTS_OBJECT_CLASS (g->node_class));
+
+    gts_container_add (GTS_CONTAINER (g), GTS_CONTAINEE (new_node));
+    if (GTS_OBJECT_CLASS (g->node_class)->read)
+      (*GTS_OBJECT_CLASS (g->node_class)->read) (&new_node, fp);
+    gts_file_first_token_after (fp, '\n');
+    nodes[n++] = GTS_GNODE (new_node);
+  }
+  if (fp->type == GTS_ERROR)
+    nn = n;
+
+  n = 0;
+  while (n < ne && fp->type != GTS_ERROR) {
+    guint n1, n2;
+
+    if (fp->type != GTS_INT)
+      gts_file_error (fp, "expecting an integer (first node index)");
+    else {
+      n1 = atoi (fp->token->str);
+      if (n1 == 0 || n1 > nn)
+	gts_file_error (fp, "node index `%d' is out of range `[1,%d]'",
+			n1, nn);
+      else {
+	gts_file_next_token (fp);
+	if (fp->type != GTS_INT)
+	  gts_file_error (fp, "expecting an integer (second node index)");
+	else {
+	  n2 = atoi (fp->token->str);
+	  if (n2 == 0 || n2 > nn)
+	    gts_file_error (fp, "node index `%d' is out of range `[1,%d]'",
+			    n2, nn);
+	  else {
+	    GtsGEdge * new_edge =
+	      gts_gedge_new (g->edge_class, nodes[n1 - 1], nodes [n2 - 1]);
+
+	    gts_file_next_token (fp);
+	    if (fp->type != '\n')
+	      if (GTS_OBJECT_CLASS (g->edge_class)->read)
+		(*GTS_OBJECT_CLASS (g->edge_class)->read)
+		  ((GtsObject **) &new_edge, fp);
+	    gts_file_first_token_after (fp, '\n');
+	    n++;
+	  }
+	}
+      }
+    }
+  }
+
+  if (fp->type == GTS_ERROR) {
+    gts_allow_floating_gnodes = TRUE;
+    while (nn)
+      gts_object_destroy (GTS_OBJECT (nodes[nn-- - 1]));
+    gts_allow_floating_gnodes = FALSE;
+  }
+  g_free (nodes);
+
+  if (fp->type == GTS_ERROR)
+    return fp->line;
+  return 0;  
+}
+
+static void write_dot_node (GtsGNode * node, gpointer * data)
+{
+  FILE * fp = data[0];
+  guint * nnode = data[1];
+
+  fprintf (fp, "  n%u", *nnode);
+  if (GTS_GNODE_CLASS (GTS_OBJECT (node)->klass)->write) {
+    fputs (" [", fp);
+    (* GTS_GNODE_CLASS (GTS_OBJECT (node)->klass)->write) (node, fp);
+    fputc (']', fp);
+  }
+  fputs (";\n", fp);
+  GTS_OBJECT (node)->reserved = GUINT_TO_POINTER ((*nnode)++);  
+}
+
+static void write_dot_edge (GtsGEdge * edge, FILE * fp)
+{
+  fprintf (fp, "  n%u -> n%u", 
+	   GPOINTER_TO_UINT (GTS_OBJECT (edge->n1)->reserved),
+	   GPOINTER_TO_UINT (GTS_OBJECT (edge->n2)->reserved));
+  if (GTS_GEDGE_CLASS (GTS_OBJECT (edge)->klass)->write) {
+    fputs (" [", fp);
+    (* GTS_GEDGE_CLASS (GTS_OBJECT (edge)->klass)->write) (edge, fp);
+    fputc (']', fp);
+  }
+  fputs (";\n", fp);
+}
+
+/**
+ * gts_graph_write_dot:
+ * @g: a #GtsGraph.
+ * @fp: a file pointer.
+ *
+ * Writes in the file @fp an ASCII representation of @g in the dot format of
+ * AT&T Bell Labs.
+ */
+void gts_graph_write_dot (GtsGraph * g, FILE * fp)
+{
+  guint nnode = 1;
+  gpointer data[2];
+
+  g_return_if_fail (g != NULL);
+  g_return_if_fail (fp != NULL);
+
+  fprintf (fp, "digraph \"%p\" {\n", g);
+  data[0] = fp;
+  data[1] = &nnode;
+  gts_container_foreach (GTS_CONTAINER (g), (GtsFunc) write_dot_node, data);
+  gts_graph_foreach_edge (g, (GtsFunc) write_dot_edge, fp);
+  fputs ("}\n", fp);
+
+  gts_container_foreach (GTS_CONTAINER (g), 
+			 (GtsFunc) gts_object_reset_reserved, NULL);
+}
+
 /* GtsWGraph */
 
 static gfloat wgraph_weight (GtsGraph * g)
@@ -1179,9 +1615,10 @@ static void create_edge (GtsEdge * e, GtsSurface * s)
       while (j) {
 	GtsFace * f1 = j->data;
 	if (GTS_IS_FACE (f1) && gts_face_has_parent_surface (f1, s))
-	  gts_gedge_new (gts_gedge_class (), 
-			 GTS_OBJECT (f)->reserved,
-			 GTS_OBJECT (f1)->reserved);
+	  gts_pgedge_new (gts_pgedge_class (), 
+			  GTS_OBJECT (f)->reserved,
+			  GTS_OBJECT (f1)->reserved,
+			  e);
 	j = j->next;
       }
     }
@@ -1214,6 +1651,53 @@ GtsGraph * gts_surface_graph_new (GtsGraphClass * klass,
   return graph;
 }
 
+static void create_segment_edge (GtsSegment * s, GtsGraph * graph)
+{
+  GtsGNode * n1 = GTS_OBJECT (s->v1)->reserved, * n2;
+
+  if (n1 == NULL) {
+    n1 = GTS_GNODE (gts_pnode_new (gts_pnode_class (), s->v1));
+    gts_container_add (GTS_CONTAINER (graph), GTS_CONTAINEE (n1));
+    GTS_OBJECT (s->v1)->reserved = n1;
+  }
+
+  n2 = GTS_OBJECT (s->v2)->reserved;
+  if (n2 == NULL) {
+    n2 = GTS_GNODE (gts_pnode_new (gts_pnode_class (), s->v2));
+    gts_container_add (GTS_CONTAINER (graph), GTS_CONTAINEE (n2));
+    GTS_OBJECT (s->v2)->reserved = n2;
+  }
+  
+  gts_pgedge_new (gts_pgedge_class (), n1, n2, s);
+}
+
+static void reset_reserved (GtsSegment * s)
+{
+  GTS_OBJECT (s->v1)->reserved = GTS_OBJECT (s->v2)->reserved = NULL;
+}
+
+/**
+ * gts_segments_graph_new:
+ * @klass: a #GtsGraphClass.
+ * @segments: a list of #GtsSegment.
+ *
+ * Returns: a new #GtsGraph representing the connectivity of the segments
+ * in @segments.
+ */
+GtsGraph * gts_segments_graph_new (GtsGraphClass * klass,
+				   GSList * segments)
+{
+  GtsGraph * graph;
+  
+  g_return_val_if_fail (klass != NULL, NULL);
+
+  graph = GTS_GRAPH (gts_object_new (GTS_OBJECT_CLASS (klass)));
+  g_slist_foreach (segments, (GFunc) create_segment_edge, graph);
+  g_slist_foreach (segments, (GFunc) reset_reserved, NULL);
+
+  return graph;
+}
+
 static void add_to_surface (GtsGNode * n, GtsSurface * s)
 {
   if (GTS_IS_FNODE (n))
