@@ -25,33 +25,31 @@
 #include "predicates.h"
 #endif /* USE_ROBUST_PREDICATES */
 
-static gint point_read (GtsObject ** o, FILE * fptr)
+static void point_read (GtsObject ** o, GtsFile * f)
 {
   GtsPoint * p = GTS_POINT (*o);
-  GString * buf;
-  gint delim;
 
-  buf = g_string_new ("");
-  delim = gts_get_token (fptr, buf, " \t\n", GTS_COMMENTS);
-  if (delim == EOF || delim == '\n') {
-    g_string_free (buf, TRUE);
-    return EOF;
+  if (f->type != GTS_INT && f->type != GTS_FLOAT) {
+    gts_file_error (f, "expecting a number (x coordinate)");
+    return;
   }
-  p->x = atof (buf->str);
-  delim = gts_get_token (fptr, buf, " \t\n", GTS_COMMENTS);
-  if (delim == EOF || delim == '\n') {
-    g_string_free (buf, TRUE);
-    return EOF;
+  p->x = atof (f->token->str);
+
+  gts_file_next_token (f);
+  if (f->type != GTS_INT && f->type != GTS_FLOAT) {
+    gts_file_error (f, "expecting a number (y coordinate)");
+    return;
   }
-  p->y = atof (buf->str);
-  delim = gts_get_token (fptr, buf, " \t\n", GTS_COMMENTS);
-  if (delim == EOF) {
-    g_string_free (buf, TRUE);
-    return EOF;
+  p->y = atof (f->token->str);
+
+  gts_file_next_token (f);
+  if (f->type != GTS_INT && f->type != GTS_FLOAT) {
+    gts_file_error (f, "expecting a number (z coordinate)");
+    return;
   }
-  p->z = atof (buf->str);
-  g_string_free (buf, TRUE);
-  return delim;
+  p->z = atof (f->token->str);
+
+  gts_file_next_token (f);
 }
 
 static void point_write (GtsObject * o, FILE * fptr)
@@ -756,6 +754,39 @@ gdouble gts_point_orientation (GtsPoint * p1, GtsPoint * p2, GtsPoint * p3)
 #endif /* USE_ROBUST_PREDICATES */
 }
 
+static gboolean ray_intersects_triangle (GtsPoint * D, GtsPoint * E,
+					 GtsTriangle * t)
+{
+  GtsPoint * A, * B, * C;
+  gint ABCE, ABCD, ADCE, ABDE, BCDE;
+
+  gts_triangle_vertices (t, (GtsVertex **) &A, 
+			 (GtsVertex **) &B, 
+			 (GtsVertex **) &C);
+
+  ABCE = gts_point_orientation_3d_sos (A, B, C, E);
+  ABCD = gts_point_orientation_3d_sos (A, B, C, D);
+  if (ABCE < 0 || ABCD > 0) {
+    GtsPoint * tmpp;
+    gint tmp;
+
+    tmpp = E; E = D; D = tmpp;
+    tmp = ABCE; ABCE = ABCD; ABCD = tmp;
+  }
+  if (ABCE < 0 || ABCD > 0)
+    return FALSE;
+  ADCE = gts_point_orientation_3d_sos (A, D, C, E);
+  if (ADCE < 0)
+    return FALSE;
+  ABDE = gts_point_orientation_3d_sos (A, B, D, E);
+  if (ABDE < 0)
+    return FALSE;
+  BCDE = gts_point_orientation_3d_sos (B, C, D, E);
+  if (BCDE < 0)
+    return FALSE;
+  return TRUE;
+}
+
 /** 
  * gts_point_is_inside_surface: 
  * @p: a #GtsPoint.  
@@ -764,33 +795,193 @@ gdouble gts_point_orientation (GtsPoint * p1, GtsPoint * p2, GtsPoint * p3)
  * @is_open: %TRUE if the surface defined by @tree is "open" i.e. its volume 
  * is negative, %FALSE otherwise.
  *
- * Returns: %TRUE if @p is inside or on the boundary of the surface
- * defined by @tree, %FALSE otherwise.  
+ * Returns: %TRUE if @p is inside the surface defined by @tree, %FALSE
+ * otherwise.  
  */
 gboolean gts_point_is_inside_surface (GtsPoint * p, 
 				      GNode * tree,
 				      gboolean is_open)
 {
-  GSList * list, * i, * stabbed = NULL;
-  gboolean inside;
+  GSList * list, * i;
+  guint nc = 0;
+  GtsPoint * p1;
+  GtsBBox * bb;
 
   g_return_val_if_fail (p != NULL, FALSE);
   g_return_val_if_fail (tree != NULL, FALSE);
 
+  bb = tree->data;
+  p1 = gts_point_new (gts_point_class (), p->x, p->y, 
+		      bb->z2 + fabs (bb->z2)/10.);
   i = list = gts_bb_tree_stabbed (tree, p);
   while (i) {
     GtsTriangle * t = GTS_TRIANGLE (GTS_BBOX (i->data)->bounded);
-    GtsObject * o = gts_triangle_is_stabbed (t, p, NULL);
-    
-    if (o && !g_slist_find (stabbed, o))
-      stabbed = g_slist_prepend (stabbed, o);
+
+    if (ray_intersects_triangle (p, p1, t))
+      nc++;
     i = i->next;
   }
   g_slist_free (list);
+  gts_object_destroy (GTS_OBJECT (p1));
 
-  inside = (g_slist_length (stabbed) % 2 != 0);
-  g_slist_free (stabbed);
-
-  return is_open ? !inside : inside;
+  return is_open ? (nc % 2 == 0) : (nc % 2 != 0);
 }
 
+#define SIGN(x) ((x) > 0. ? 1 : -1)
+#define ORIENT1D(a,b) ((a) > (b) ? 1 : (a) < (b) ? -1 : 0)
+
+static gint sortp (gpointer * p, guint n)
+{
+  gint sign = 1;
+  guint i, j;
+
+  for (i = 0; i < n - 1; i++)
+    for (j = 0; j < n - 1 - i; j++)
+      if (GPOINTER_TO_UINT (p[j+1]) < GPOINTER_TO_UINT (p[j])) {
+	gpointer tmp = p[j];
+
+	p[j] = p[j+1];
+	p[j+1] = tmp;
+	sign = - sign;
+      }
+  return sign;
+}
+
+/**
+ * gts_point_orientation_3d_sos:
+ * @p1: a #GtsPoint.
+ * @p2: a #GtsPoint.
+ * @p3: a #GtsPoint.
+ * @p4: a #GtsPoint.
+ *
+ * Checks if @p4 lies above or below the plane passing through the
+ * points @p1, @p2 and @p3. Below is defined so that @p1, @p2 and @p3
+ * appear in counterclockwise order when viewed from above the
+ * plane. This function uses adaptive floating point arithmetic and is
+ * consequently geometrically robust.
+ *
+ * Simulation of Simplicity (SoS) is used to break ties when the
+ * orientation is degenerate (i.e. @p4 lies on the plane defined by
+ * @p1, @p2 and @p3).
+ *
+ * Returns: +1 if @p4 lies below, -1 if @p4 lies above the plane.  
+ */
+gint gts_point_orientation_3d_sos (GtsPoint * p1,
+				   GtsPoint * p2,
+				   GtsPoint * p3,
+				   GtsPoint * p4)
+{
+#ifndef USE_ROBUST_PREDICATES
+  g_assert_not_reached ();
+  return 0;
+#else /* USE_ROBUST_PREDICATES */
+  gdouble o;
+
+  g_return_val_if_fail (p1 != NULL && p2 != NULL && 
+			p3 != NULL && p4 != NULL, 0);
+
+  o = orient3d ((gdouble *) &p1->x, 
+		(gdouble *) &p2->x, 
+		(gdouble *) &p3->x, 
+		(gdouble *) &p4->x);
+  if (o != 0.)
+    return SIGN (o);
+  else {
+    GtsPoint * p[4];
+    gdouble a[2], b[2], c[2];
+    gint sign;
+
+    p[0] = p1; p[1] = p2; p[2] = p3; p[3] = p4;
+    sign = sortp ((gpointer *) p, 4);
+    
+    /* epsilon^1/8 */
+    a[0] = p[1]->x; a[1] = p[1]->y;
+    b[0] = p[2]->x; b[1] = p[2]->y;
+    c[0] = p[3]->x; c[1] = p[3]->y;
+    o = orient2d (a, b, c);
+    if (o != 0.)
+      return SIGN (o)*sign;
+    
+    /* epsilon^1/4 */
+    a[0] = p[1]->x; a[1] = p[1]->z;
+    b[0] = p[2]->x; b[1] = p[2]->z;
+    c[0] = p[3]->x; c[1] = p[3]->z;
+    o = orient2d (a, b, c);
+    if (o != 0.)
+      return - SIGN (o)*sign;
+    
+    /* epsilon^1/2 */
+    a[0] = p[1]->y; a[1] = p[1]->z;
+    b[0] = p[2]->y; b[1] = p[2]->z;
+    c[0] = p[3]->y; c[1] = p[3]->z;
+    o = orient2d (a, b, c);
+    if (o != 0.)
+      return SIGN (o)*sign;
+    
+    /* epsilon */
+    a[0] = p[0]->x; a[1] = p[0]->y;
+    b[0] = p[2]->x; b[1] = p[2]->y;
+    c[0] = p[3]->x; c[1] = p[3]->y;
+    o = orient2d (a, b, c);
+    if (o != 0.)
+      return - SIGN (o)*sign;
+    
+    /* epsilon^5/4 */
+    o = ORIENT1D (p[2]->x, p[3]->x);
+    if (o != 0.)
+      return SIGN (o)*sign;
+    
+    /* epsilon^3/2 */
+    o = ORIENT1D (p[2]->y, p[3]->y);
+    if (o != 0.)
+      return - SIGN (o)*sign;
+    
+    /* epsilon^2 */
+    a[0] = p[0]->x; a[1] = p[0]->z;
+    b[0] = p[2]->x; b[1] = p[2]->z;
+    c[0] = p[3]->x; c[1] = p[3]->z;
+    o = orient2d (a, b, c);
+    if (o != 0.)
+      return SIGN (o)*sign;
+
+    /* epsilon^5/2 */
+    o = ORIENT1D (p[2]->z, p[3]->z);
+    if (o != 0.)
+      return SIGN (o)*sign;
+    
+    /* epsilon^4 */
+    a[0] = p[0]->y; a[1] = p[0]->z;
+    b[0] = p[2]->y; b[1] = p[2]->z;
+    c[0] = p[3]->y; c[1] = p[3]->z;
+    o = orient2d (a, b, c);
+    if (o != 0.)
+      return - SIGN (o)*sign;
+
+    /* epsilon^8 */
+    a[0] = p[0]->x; a[1] = p[0]->y;
+    b[0] = p[1]->x; b[1] = p[1]->y;
+    c[0] = p[3]->x; c[1] = p[3]->y;
+    o = orient2d (a, b, c);
+    if (o != 0.)
+      return SIGN (o)*sign;
+
+    /* epsilon^33/4 */
+    o = ORIENT1D (p[1]->x, p[3]->x);
+    if (o != 0.)
+      return - SIGN (o)*sign;
+    
+    /* epsilon^17/2 */
+    o = ORIENT1D (p[1]->y, p[3]->y);
+    if (o != 0.)
+      return SIGN (o)*sign;
+    
+    /* epsilon^10 */
+    o = ORIENT1D (p[0]->x, p[3]->x);
+    if (o != 0.)
+      return SIGN (o)*sign;
+    
+    /* epsilon^21/2 */
+    return sign;
+  }
+#endif /* USE_ROBUST_PREDICATES */
+}
