@@ -18,6 +18,7 @@
  */
 
 #include <stdlib.h>
+#include <string.h>
 #include "gts.h"
 
 #define DYNAMIC_SPLIT
@@ -1330,15 +1331,15 @@ void gts_psurface_write (GtsPSurface * ps, FILE * fptr)
 
       a = scf->a1;
       while ((t = *(a++)))
-	fprintf (fptr, "%u\n",
+	fprintf (fptr, "%u ",
 		 GPOINTER_TO_UINT (g_hash_table_lookup (hash, t)));
-      fprintf (fptr, "0\n");
+      fprintf (fptr, "\n");
 
       a = scf->a2;
       while ((t = *(a++)))
-	fprintf (fptr, "%u\n",
+	fprintf (fptr, "%u ",
 		 GPOINTER_TO_UINT (g_hash_table_lookup (hash, t)));
-      fprintf (fptr, "0\n");
+      fprintf (fptr, "\n");
 
       g_hash_table_insert (hash, cf, GUINT_TO_POINTER (nf++));
 
@@ -1353,160 +1354,161 @@ void gts_psurface_write (GtsPSurface * ps, FILE * fptr)
   g_hash_table_destroy (hash);
 }
 
-static GtsSurface * surface_read (GtsSurface * surface, 
-				  FILE * fptr, 
-				  guint * line,
-				  GPtrArray * vertices,
-				  GPtrArray * faces)
+static guint surface_read (GtsSurface * surface, 
+			   GtsFile * f,
+			   GPtrArray * vertices,
+			   GPtrArray * faces)
 {
   GtsEdge ** edges;
-  guint n, np, ns, nt, line_number = 1;
-  gboolean syntax_error = FALSE;
-  GString * buf;
-  gint delim;
+  guint n, nv, ne, nf;
 
-  g_return_val_if_fail (surface != NULL, NULL);
-  g_return_val_if_fail (fptr != NULL, NULL);
-
-  buf = g_string_new ("");
-
-  delim = gts_get_token (fptr, buf, " \t\n", GTS_COMMENTS);
-  if (delim == EOF || delim == '\n') {
-    g_string_free (buf, TRUE);
-    return NULL;
+  g_return_val_if_fail (surface != NULL, 1);
+  g_return_val_if_fail (f != NULL, 1);
+  
+  if (f->type != GTS_INT) {
+    gts_file_error (f, "expecting an integer (number of vertices)");
+    return f->line;
   }
-  np = atoi (buf->str);
-  delim = gts_get_token (fptr, buf, " \t\n", GTS_COMMENTS);
-  if (delim == EOF || delim == '\n') {
-    g_string_free (buf, TRUE);
-    return NULL;
-  }
-  ns = atoi (buf->str);
-  delim = gts_get_token (fptr, buf, " \t\n", GTS_COMMENTS);
-  if (delim == EOF) {
-    g_string_free (buf, TRUE);
-    return NULL;
-  }
-  nt = atoi (buf->str);
-  if (delim != '\n')
-    gts_get_newline (fptr);
+  nv = atoi (f->token->str);
 
-  g_ptr_array_set_size (vertices, np);
-  g_ptr_array_set_size (faces, nt);
-  /* allocate ns + 1 just in case ns == 0 */
-  edges = g_malloc ((ns + 1)*sizeof (GtsEdge *));
+  gts_file_next_token (f);
+  if (f->type != GTS_INT) {
+    gts_file_error (f, "expecting an integer (number of edges)");
+    return f->line;
+  }
+  ne = atoi (f->token->str);
+
+  gts_file_next_token (f);
+  if (f->type != GTS_INT) {
+    gts_file_error (f, "expecting an integer (number of faces)");
+    return f->line;
+  }
+  nf = atoi (f->token->str);
+
+  gts_file_first_token_after (f, '\n');
+
+  g_ptr_array_set_size (vertices, nv);
+  g_ptr_array_set_size (faces, nf);
+  /* allocate nv + 1 just in case nv == 0 */
+  edges = g_malloc ((ne + 1)*sizeof (GtsEdge *));
   
   n = 0;
-  while (n < np && !syntax_error) {
-    GtsObject * new_vertex = 
+  while (n < nv && f->type != GTS_ERROR) {
+    GtsObject * new_vertex =
       gts_object_new (GTS_OBJECT_CLASS (surface->vertex_class));
 
-    line_number++;
-    delim = (*GTS_OBJECT_CLASS (surface->vertex_class)->read)
-      (&new_vertex, fptr);
-    if (delim == EOF)
-      syntax_error = TRUE;
-    else if (delim != '\n')
-      gts_get_newline (fptr);
-    g_ptr_array_index (vertices, n++) = new_vertex;
+    (* GTS_OBJECT_CLASS (surface->vertex_class)->read) (&new_vertex, f);
+    if (f->type != GTS_ERROR) {
+      gts_file_first_token_after (f, '\n');
+      g_ptr_array_index (vertices, n++) = new_vertex;
+    }
   }
-  if (syntax_error)
-    np = n;
+  if (f->type == GTS_ERROR)
+    nv = n;
 
   n = 0;
-  while (n < ns && !syntax_error) {
+  while (n < ne && f->type != GTS_ERROR) {
     guint p1, p2;
 
-    line_number++;
-    delim = gts_get_token (fptr, buf, " \t\n", GTS_COMMENTS);
-    if (delim == EOF || delim == '\n')
-      syntax_error = TRUE;
+    if (f->type != GTS_INT)
+      gts_file_error (f, "expecting an integer (first vertex index)");
     else {
-      p1 = atoi (buf->str);
-      delim = gts_get_token (fptr, buf, " \t\n", GTS_COMMENTS);
-      if (delim == EOF)
-	syntax_error = TRUE;
+      p1 = atoi (f->token->str);
+      if (p1 == 0 || p1 > nv)
+	gts_file_error (f, "vertex index `%d' is out of range `[1,%d]'", 
+			p1, nv);
       else {
-	p2 = atoi (buf->str);
-	if (p1 == 0 || p2 == 0 || p1 > np || p2 > np)
-	  syntax_error = TRUE;
+	gts_file_next_token (f);
+	if (f->type != GTS_INT)
+	  gts_file_error (f, "expecting an integer (second vertex index)");
 	else {
-	  GtsEdge * new_edge = 
-	    gts_edge_new (surface->edge_class,
-			  g_ptr_array_index (vertices, p1 - 1),
-			  g_ptr_array_index (vertices, p2 - 1));
-	  if (delim != '\n') {
-	    if (GTS_OBJECT_CLASS (surface->edge_class)->read)
-	      delim = (*GTS_OBJECT_CLASS (surface->edge_class)->read)
-		((GtsObject **) &new_edge, fptr);
-	    if (delim != '\n')
-	      gts_get_newline (fptr);
+	  p2 = atoi (f->token->str);
+	  if (p2 == 0 || p2 > nv)
+	    gts_file_error (f, "vertex index `%d' is out of range `[1,%d]'", 
+			    p2, nv);
+	  else {
+	    GtsEdge * new_edge =
+	      gts_edge_new (surface->edge_class,
+			    g_ptr_array_index (vertices, p1 - 1),
+			    g_ptr_array_index (vertices, p2 - 1));
+
+	    gts_file_next_token (f);
+	    if (f->type != '\n')
+	      if (GTS_OBJECT_CLASS (surface->edge_class)->read)
+		(*GTS_OBJECT_CLASS (surface->edge_class)->read)
+		  ((GtsObject **) &new_edge, f);
+	    gts_file_first_token_after (f, '\n');
+	    edges[n++] = new_edge;
 	  }
-	  edges[n++] = new_edge;
 	}
       }
     }
   }
-  if (syntax_error)
-    ns = n;
+  if (f->type == GTS_ERROR)
+    ne = n;
 
   n = 0;
-  while (n < nt && !syntax_error) {
+  while (n < nf && f->type != GTS_ERROR) {
     guint s1, s2, s3;
 
-    line_number++;
-    delim = gts_get_token (fptr, buf, " \t\n", GTS_COMMENTS);
-    if (delim == EOF || delim == '\n')
-      syntax_error = TRUE;
+    if (f->type != GTS_INT)
+      gts_file_error (f, "expecting an integer (first edge index)");
     else {
-      s1 = atoi (buf->str);
-      delim = gts_get_token (fptr, buf, " \t\n", GTS_COMMENTS);
-      if (delim == EOF || delim == '\n')
-	syntax_error = TRUE;
+      s1 = atoi (f->token->str);
+      if (s1 == 0 || s1 > ne)
+	gts_file_error (f, "edge index `%d' is out of range `[1,%d]'", 
+			s1, ne);
       else {
-	s2 = atoi (buf->str);
-	delim = gts_get_token (fptr, buf, " \t\n", GTS_COMMENTS);
-	if (delim == EOF)
-	  syntax_error = TRUE;
+	gts_file_next_token (f);
+	if (f->type != GTS_INT)
+	  gts_file_error (f, "expecting an integer (second edge index)");
 	else {
-	  s3 = atoi (buf->str);
-	  if (s1 == 0 || s2 == 0 || s3 == 0 || s1 > ns || s2 > ns || s3 > ns)
-	    syntax_error = TRUE;
+	  s2 = atoi (f->token->str);
+	  if (s2 == 0 || s2 > ne)
+	    gts_file_error (f, "edge index `%d' is out of range `[1,%d]'", 
+			    s2, ne);
 	  else {
-	    GtsFace * new_face = gts_face_new (surface->face_class,
-					       edges[s1 - 1],
-					       edges[s2 - 1],
-					       edges[s3 - 1]);
-	    if (delim != '\n') {
-	      if (GTS_OBJECT_CLASS (surface->face_class)->read)
-		delim = (*GTS_OBJECT_CLASS (surface->face_class)->read)
-		  ((GtsObject **) &new_face, fptr);
-	      if (delim != '\n')
-		gts_get_newline (fptr);
+	    gts_file_next_token (f);
+	    if (f->type != GTS_INT)
+	      gts_file_error (f, "expecting an integer (third edge index)");
+	    else {
+	      s3 = atoi (f->token->str);
+	      if (s3 == 0 || s3 > ne)
+		gts_file_error (f, "edge index `%d' is out of range `[1,%d]'", 
+				s3, ne);
+	      else {
+		GtsFace * new_face = gts_face_new (surface->face_class,
+						   edges[s1 - 1],
+						   edges[s2 - 1],
+						   edges[s3 - 1]);
+
+		gts_file_next_token (f);
+		if (f->type != '\n')
+		  if (GTS_OBJECT_CLASS (surface->face_class)->read)
+		    (*GTS_OBJECT_CLASS (surface->face_class)->read)
+		      ((GtsObject **) &new_face, f);
+		gts_file_first_token_after (f, '\n');
+		gts_surface_add_face (surface, new_face);
+		g_ptr_array_index (faces, n++) = new_face;
+	      }
 	    }
-	    gts_surface_add_face (surface, new_face);
-	    g_ptr_array_index (faces, n++) = new_face;
 	  }
 	}
       }
     }
   }
 
-  if (syntax_error) {
-    while (np)
-      gts_object_destroy (GTS_OBJECT (g_ptr_array_index (vertices, np-- - 1)));
-    while (ns)
-      gts_object_destroy (GTS_OBJECT (edges[ns-- - 1]));
-    gts_object_destroy (GTS_OBJECT (surface));
-    surface = NULL;
-    if (line)
-      *line = line_number;
-  }
-
   g_free (edges);
-  g_string_free (buf, TRUE);
 
-  return surface;
+  if (f->type == GTS_ERROR) {
+    gts_allow_floating_vertices = TRUE;
+    while (nv)
+      gts_object_destroy (GTS_OBJECT (g_ptr_array_index (vertices, nv-- - 1)));
+    gts_allow_floating_vertices = FALSE;    
+    return f->line;
+  }
+
+  return 0;
 }
 
 /**
@@ -1514,11 +1516,9 @@ static GtsSurface * surface_read (GtsSurface * surface,
  * @klass: a #GtsPSurfaceClass.
  * @s: a #GtsSurface.
  * @split_class: a #GtsSplitClass to use for the #GtsSplit.
- * @fptr: a file pointer.
- * @line: %NULL or a pointer to a variable to fill with the line number in case
- * of an error in the file format. 
+ * @f: a #GtsFile.
  *
- * Creates a new #GtsPSurface prepared for input from the file @fptr 
+ * Creates a new #GtsPSurface prepared for input from the file @f 
  * containing a valid GTS representation of a progressive surface. The initial
  * shape of the progressive surface is loaded into @s.
  * 
@@ -1528,22 +1528,19 @@ static GtsSurface * surface_read (GtsSurface * surface,
  * gts_psurface_max_vertex_number() can still be used.
  *
  * Returns: a new #GtsPSurface or %NULL if there was a format error while
- * reading the file, in which case @line is set to the line where
- * the error occured.
+ * reading the file, in which case @f contains information about the error.
  */
 GtsPSurface * gts_psurface_open (GtsPSurfaceClass * klass,
 				 GtsSurface * s,
 				 GtsSplitClass * split_class,
-				 FILE * fptr,
-				 guint * line)
+				 GtsFile * f)
 {
   GtsPSurface * ps;
-  guint ns = 0;
 
   g_return_val_if_fail (klass != NULL, NULL);
   g_return_val_if_fail (s != NULL, NULL);
   g_return_val_if_fail (split_class != NULL, NULL);
-  g_return_val_if_fail (fptr != NULL, NULL);
+  g_return_val_if_fail (f != NULL, NULL);
 
   ps = GTS_PSURFACE (gts_object_new (GTS_OBJECT_CLASS (klass)));
   ps->s = s;
@@ -1552,14 +1549,23 @@ GtsPSurface * gts_psurface_open (GtsPSurfaceClass * klass,
   ps->vertices = g_ptr_array_new ();
   ps->faces = g_ptr_array_new ();
 
-  if (surface_read (s, fptr, line, ps->vertices, ps->faces) == NULL)
-    return ps;
+  if (surface_read (s, f, ps->vertices, ps->faces)) {
+    ps->s = NULL;
+    gts_object_destroy (GTS_OBJECT (ps));
+    return NULL;
+  }
 
   ps->min = gts_surface_vertex_number (ps->s);
   ps->pos = 0;
 
-  if (fscanf (fptr, "%u", &ns) == 1)
-    g_ptr_array_set_size (ps->split, ns);
+  if (f->type == GTS_INT) {
+    gint ns = atoi (f->token->str);
+    
+    if (ns > 0) {
+      g_ptr_array_set_size (ps->split, ns);
+      gts_file_first_token_after (f, '\n');
+    }
+  }
 
   return ps;
 }
@@ -1567,46 +1573,47 @@ GtsPSurface * gts_psurface_open (GtsPSurfaceClass * klass,
 /**
  * gts_psurface_read_vertex:
  * @ps: a #GtsPSurface prealably created with gts_psurface_open().
- * @fptr: a file pointer.
+ * @fp: a #GtsFile.
  *
- * Reads in one vertex split operation from @fptr and performs the expansion.
+ * Reads in one vertex split operation from @fp and performs the expansion.
+ *
+ * If an error occurs while reading the file, the @error field of @fp is set.
  *
  * Returns: the newly created #GtsSplit or %NULL if no vertex split could be
- * read from @fptr.
+ * read from @fp.
  */
-GtsSplit * gts_psurface_read_vertex (GtsPSurface * ps, FILE * fptr)
+GtsSplit * gts_psurface_read_vertex (GtsPSurface * ps, GtsFile * fp)
 {
   guint nv, ncf;
   GtsSplit * vs, * parent;
   GtsSplitCFace * scf;
-  gboolean ok = TRUE;
-  gint delim;
-  GString * buf;
 
   g_return_val_if_fail (ps != NULL, NULL);
-  g_return_val_if_fail (fptr != NULL, NULL);
+  g_return_val_if_fail (fp != NULL, NULL);
   g_return_val_if_fail (!GTS_PSURFACE_IS_CLOSED (ps), NULL);
   
   if (ps->pos >= ps->split->len)
     return NULL;
 
-  buf = g_string_new ("");
-  delim = gts_get_token (fptr, buf, " \t\n", GTS_COMMENTS);
-  if (delim == EOF || delim == '\n') {
-    g_string_free (buf, TRUE);
+  if (fp->type == GTS_NONE)
+    return NULL;
+  if (fp->type != GTS_INT) {
+    gts_file_error (fp, "expecting an integer (vertex index)");
     return NULL;
   }
-  nv = atoi (buf->str);
+  nv = atoi (fp->token->str);
   if (nv == 0 || nv > ps->vertices->len) {
-    g_string_free (buf, TRUE);
+    gts_file_error (fp, "vertex index `%d' is out of range `[1,%d]'",
+		    nv, ps->vertices->len);
     return NULL;
   }
-  delim = gts_get_token (fptr, buf, " \t\n", GTS_COMMENTS);
-  if (delim == EOF) {
-    g_string_free (buf, TRUE);
+
+  gts_file_next_token (fp);
+  if (fp->type != GTS_INT) {
+    gts_file_error (fp, "expecting an integer (ncf)");
     return NULL;
   }
-  ncf = atoi (buf->str);
+  ncf = atoi (fp->token->str);
   
   vs = GTS_SPLIT (gts_object_new (GTS_OBJECT_CLASS (ps->split_class)));
 
@@ -1615,107 +1622,127 @@ GtsSplit * gts_psurface_read_vertex (GtsPSurface * ps, FILE * fptr)
   vs->cfaces = NULL;
   vs->ncf = 0;
 
-  if (delim != '\n') {
+  gts_file_next_token (fp);
+  if (fp->type != '\n')
     if (GTS_OBJECT (vs)->klass->read)
-      delim = (*GTS_OBJECT (vs)->klass->read) ((GtsObject **) &vs, fptr);
-    if (delim != '\n')
-      gts_get_newline (fptr);
-  }
-  
-  vs->v1 = gts_object_new (GTS_OBJECT_CLASS (ps->s->vertex_class));
+      (* GTS_OBJECT (vs)->klass->read) ((GtsObject **) &vs, fp);
+  gts_file_first_token_after (fp, '\n');
 
-  delim = (*GTS_OBJECT_CLASS (ps->s->vertex_class)->read) 
-    (&(vs->v1), fptr);
-  if (delim == EOF)
-    ok = FALSE;
-  else {
-    vs->v1->reserved = vs;
-    g_ptr_array_add (ps->vertices, vs->v1);
+  if (fp->type != GTS_ERROR) {
+    vs->v1 = gts_object_new (GTS_OBJECT_CLASS (ps->s->vertex_class));
+    (* GTS_OBJECT_CLASS (ps->s->vertex_class)->read) (&(vs->v1), fp);
+    if (fp->type != GTS_ERROR) {
+      vs->v1->reserved = vs;
+      g_ptr_array_add (ps->vertices, vs->v1);
 
-    if (delim != '\n') 
-      gts_get_newline (fptr);
-
-    vs->v2 = gts_object_new (GTS_OBJECT_CLASS (ps->s->vertex_class));
-
-    delim = (*GTS_OBJECT_CLASS (ps->s->vertex_class)->read) 
-      (&(vs->v2), fptr);
-    if (delim == EOF)
-      ok = FALSE;
-    else {
-      vs->v2->reserved = vs;
-      g_ptr_array_add (ps->vertices, vs->v2);
-
-      if (delim != '\n')
-	gts_get_newline (fptr);
+      gts_file_first_token_after (fp, '\n');
+      
+      vs->v2 = gts_object_new (GTS_OBJECT_CLASS (ps->s->vertex_class));
+      (*GTS_OBJECT_CLASS (ps->s->vertex_class)->read) (&(vs->v2), fp);
+      if (fp->type != GTS_ERROR) {
+	vs->v2->reserved = vs;
+	g_ptr_array_add (ps->vertices, vs->v2);
+	gts_file_first_token_after (fp, '\n');
+      }
     }
   }
-  if (ok) {
+
+  if (fp->type != GTS_ERROR) {
     scf = vs->cfaces = g_malloc (sizeof (GtsSplitCFace)*ncf);
-    while (ok && ncf--) {
+    while (fp->type != GTS_ERROR && ncf--) {
       guint it, flags;
       GtsFace * f;
       CFace * cf;
       GPtrArray * a;
 
-      delim = gts_get_token (fptr, buf, " \t\n", GTS_COMMENTS);
-      if (delim == EOF || delim == '\n')
-	ok = FALSE;
+      if (fp->type != GTS_INT)
+	gts_file_error (fp, "expecting an integer (face index)");
       else {
-	it = atoi (buf->str);
-	delim = gts_get_token (fptr, buf, " \t\n", GTS_COMMENTS);
-	if (delim == EOF || it == 0 || it > ps->faces->len)
-	  ok = FALSE;
+	it = atoi (fp->token->str);
+	if (it == 0 || it > ps->faces->len)
+	  gts_file_error (fp, "face index `%d' is out of range `[1,%d]'",
+			  it, ps->faces->len);
 	else {
-	  flags = atoi (buf->str);
-	  f = GTS_FACE (gts_object_new (GTS_OBJECT_CLASS (ps->s->face_class)));
-	  
-	  if (delim != '\n') {
-	    if (GTS_OBJECT (f)->klass->read)
-	      delim = (*GTS_OBJECT (f)->klass->read) 
-		((GtsObject **) &f, fptr);
-	    if (delim != '\n')
-	      gts_get_newline (fptr);
-	  }
-	  scf->f = f;
+	  gts_file_next_token (fp);
+	  if (fp->type != GTS_INT)
+	    gts_file_error (fp, "expecting an integer (flags)");
+	  else {
+	    flags = atoi (fp->token->str);
+	    f = 
+	      GTS_FACE (gts_object_new (GTS_OBJECT_CLASS (ps->s->face_class)));
 
-	  cf = (CFace *) f;
-	  GTS_OBJECT (cf)->klass = GTS_OBJECT_CLASS (cface_class ());
-	  cf->parent_split = vs;
-	  cf->t = g_ptr_array_index (ps->faces, it - 1);
-	  cf->flags = flags;
+	    gts_file_next_token (fp);
+	    if (fp->type != '\n')
+	      if (GTS_OBJECT (f)->klass->read)
+		(*GTS_OBJECT (f)->klass->read) ((GtsObject **) &f, fp);
+	    gts_file_first_token_after (fp, '\n');
+	    if (fp->type != GTS_ERROR) {
+	      scf->f = f;
+
+	      cf = (CFace *) f;
+	      GTS_OBJECT (cf)->klass = GTS_OBJECT_CLASS (cface_class ());
+	      cf->parent_split = vs;
+	      cf->t = g_ptr_array_index (ps->faces, it - 1);
+	      cf->flags = flags;
 	  
-	  a = g_ptr_array_new ();
-	  while (ok && fscanf (fptr, "%u", &it) == 1 && it > 0) {
-	    if (it > ps->faces->len)
-	      ok = FALSE;
-	    else
-	      g_ptr_array_add (a, g_ptr_array_index (ps->faces, it - 1));
+	      a = g_ptr_array_new ();
+	      do {
+		if (fp->type != GTS_INT)
+		  gts_file_error (fp, "expecting an integer (face index)");
+		else {
+		  it = atoi (fp->token->str);
+		  if (it > ps->faces->len)
+		    gts_file_error (fp, 
+				    "face index `%d' is out of range `[1,%d]'",
+				    it, ps->faces->len);
+		  else {
+		    g_ptr_array_add (a, g_ptr_array_index (ps->faces, 
+							   it - 1));
+		    gts_file_next_token (fp);
+		  }
+		}
+	      } while (fp->type != GTS_ERROR && fp->type != '\n');
+	      gts_file_first_token_after (fp, '\n');
+	      g_ptr_array_add (a, NULL);
+	      scf->a1 = (GtsTriangle **) a->pdata;
+	      g_ptr_array_free (a, FALSE);
+
+	      if (fp->type != GTS_ERROR) {
+		a = g_ptr_array_new ();
+		do {
+		  if (fp->type != GTS_INT)
+		    gts_file_error (fp, "expecting an integer (face index)");
+		  else {
+		    it = atoi (fp->token->str);
+		    if (it > ps->faces->len)
+		      gts_file_error (fp, 
+				   "face index `%d' is out of range `[1,%d]'",
+				      it, ps->faces->len);
+		    else {
+		      g_ptr_array_add (a, g_ptr_array_index (ps->faces, 
+							     it - 1));
+		      gts_file_next_token (fp);
+		    }
+		  }
+		} while (fp->type != GTS_ERROR && fp->type != '\n');
+		gts_file_first_token_after (fp, '\n');
+		g_ptr_array_add (a, NULL);
+		scf->a2 = (GtsTriangle **) a->pdata;
+		g_ptr_array_free (a, FALSE);
+		
+		g_ptr_array_add (ps->faces, f);
+	      
+		vs->ncf++;
+		scf++;
+	      }
+	    }
 	  }
-	  g_ptr_array_add (a, NULL);
-	  scf->a1 = (GtsTriangle **) a->pdata;
-	  g_ptr_array_free (a, FALSE);
-	  
-	  a = g_ptr_array_new ();
-	  while (ok && fscanf (fptr, "%u", &it) == 1 && it > 0) {
-	    if (it > ps->faces->len)
-	      ok = FALSE;
-	    else
-	      g_ptr_array_add (a, g_ptr_array_index (ps->faces, it - 1));
-	  }
-	  g_ptr_array_add (a, NULL);
-	  scf->a2 = (GtsTriangle **) a->pdata;
-	  g_ptr_array_free (a, FALSE);
-	  
-	  g_ptr_array_add (ps->faces, f);
-	  
-	  vs->ncf++;
-	  scf++;
 	}
       }
     }
   }
 
-  if (ok) {
+  if (fp->type != GTS_ERROR) {
     if ((parent = GTS_OBJECT (vs->v)->reserved)) {
       GTS_OBJECT (vs->v)->reserved = NULL;
       if (parent->v1 == GTS_OBJECT (vs->v))
@@ -1727,8 +1754,6 @@ GtsSplit * gts_psurface_read_vertex (GtsPSurface * ps, FILE * fptr)
     }
     g_ptr_array_index (ps->split, ps->pos++) = vs;
     gts_split_expand (vs, ps->s, ps->s->edge_class);
-    
-    g_string_free (buf, TRUE);
 
     return vs;
   }
@@ -1737,7 +1762,6 @@ GtsSplit * gts_psurface_read_vertex (GtsPSurface * ps, FILE * fptr)
   if (vs->v2) gts_object_destroy (vs->v2);
   gts_object_destroy (GTS_OBJECT (vs));
   
-  g_string_free (buf, TRUE);
   return NULL;
 }
 
