@@ -21,6 +21,8 @@
 #include <math.h>
 #include "gts.h"
 
+#include "gts-private.h"
+
 static void destroy_foreach_face (GtsFace * f, GtsSurface * s)
 {
   f->surfaces = g_slist_remove (f->surfaces, s);
@@ -34,7 +36,11 @@ static void surface_destroy (GtsObject * object)
   GtsSurface * surface = GTS_SURFACE (object);
   
   gts_surface_foreach_face (surface, (GtsFunc) destroy_foreach_face, surface);
+#ifdef USE_SURFACE_BTREE
+  g_tree_destroy (surface->faces);
+#else /* not USE_SURFACE_BTREE */
   g_hash_table_destroy (surface->faces);
+#endif /* not USE_SURFACE_BTREE */
 
   (* GTS_OBJECT_CLASS (gts_surface_class ())->parent_class->destroy) (object);
 }
@@ -56,9 +62,24 @@ static void surface_class_init (GtsSurfaceClass * klass)
   klass->remove_face = NULL;
 }
 
+#ifdef USE_SURFACE_BTREE
+static gint compare_pointers (gconstpointer a, gconstpointer b)
+{
+  if (GPOINTER_TO_UINT (a) < GPOINTER_TO_UINT (b))
+    return -1;
+  if (GPOINTER_TO_UINT (a) > GPOINTER_TO_UINT (b))
+    return 1;
+  return 0;
+}
+#endif /* USE_SURFACE_BTREE */
+
 static void surface_init (GtsSurface * surface)
 {
+#ifdef USE_SURFACE_BTREE
+  surface->faces = g_tree_new (compare_pointers);
+#else /* not USE_SURFACE_BTREE */
   surface->faces = g_hash_table_new (NULL, NULL);
+#endif /* not USE_SURFACE_BTREE */
   surface->vertex_class = gts_vertex_class ();
   surface->edge_class = gts_edge_class ();
   surface->face_class = gts_face_class ();
@@ -128,10 +149,17 @@ void gts_surface_add_face (GtsSurface * s, GtsFace * f)
 
   g_assert (s->keep_faces == FALSE);
 
+#ifdef USE_SURFACE_BTREE
+  if (!g_tree_lookup (s->faces, f)) {
+    f->surfaces = g_slist_prepend (f->surfaces, s);
+    g_tree_insert (s->faces, f, f);
+  }
+#else /* not USE_SURFACE_BTREE */
   if (!g_hash_table_lookup (s->faces, f)) {
     f->surfaces = g_slist_prepend (f->surfaces, s);
     g_hash_table_insert (s->faces, f, f);
   }
+#endif /* not USE_SURFACE_BTREE */
 
   if (GTS_SURFACE_CLASS (GTS_OBJECT (s)->klass)->add_face)
     (* GTS_SURFACE_CLASS (GTS_OBJECT (s)->klass)->add_face) (s, f);
@@ -152,7 +180,12 @@ void gts_surface_remove_face (GtsSurface * s,
 
   g_assert (s->keep_faces == FALSE);
 
+#ifdef USE_SURFACE_BTREE
+  g_tree_remove (s->faces, f);
+#else /* not USE_SURFACE_BTREE */
   g_hash_table_remove (s->faces, f);
+#endif /* not USE_SURFACE_BTREE */
+
   f->surfaces = g_slist_remove (f->surfaces, s);
 
   if (GTS_SURFACE_CLASS (GTS_OBJECT (s)->klass)->remove_face)
@@ -167,167 +200,171 @@ void gts_surface_remove_face (GtsSurface * s,
 /**
  * gts_surface_read:
  * @surface: a #GtsSurface.
- * @fptr: a file pointer.
+ * @f: a #GtsFile.
  *
- * Add to @surface the data read from @fptr. The format of the file pointed to
- * by @fptr is as described in gts_surface_write().
+ * Add to @surface the data read from @f. The format of the file pointed to
+ * by @f is as described in gts_surface_write().
  *
- * Returns: 0 if successful or the line number at which the parsing stopped in
- * case of error.
+ * Returns: 0 if successful or the line number at which the parsing
+ * stopped in case of error (in which case the @error field of @f is
+ * set to a description of the error which occured).  
  */
-guint gts_surface_read (GtsSurface * surface, FILE * fptr)
+/* Update split.c/surface_read() if modifying this function */
+guint gts_surface_read (GtsSurface * surface, GtsFile * f)
 {
   GtsVertex ** vertices;
   GtsEdge ** edges;
-  guint n, nv, ne, nf, line_number = 1;
-  gboolean syntax_error = FALSE;
-  GString * buf;
-  gint delim;
+  guint n, nv, ne, nf;
 
   g_return_val_if_fail (surface != NULL, 1);
-  g_return_val_if_fail (fptr != NULL, 1);
+  g_return_val_if_fail (f != NULL, 1);
 
-  buf = g_string_new ("");
+  if (f->type != GTS_INT) {
+    gts_file_error (f, "expecting an integer (number of vertices)");
+    return f->line;
+  }
+  nv = atoi (f->token->str);
 
-  delim = gts_get_token (fptr, buf, " \t\n", GTS_COMMENTS);
-  if (delim == EOF || delim == '\n') {
-    g_string_free (buf, TRUE);
-    return 1;
+  gts_file_next_token (f);
+  if (f->type != GTS_INT) {
+    gts_file_error (f, "expecting an integer (number of edges)");
+    return f->line;
   }
-  nv = atoi (buf->str);
-  delim = gts_get_token (fptr, buf, " \t\n", GTS_COMMENTS);
-  if (delim == EOF || delim == '\n') {
-    g_string_free (buf, TRUE);
-    return 1;
-  }
-  ne = atoi (buf->str);
-  delim = gts_get_token (fptr, buf, " \t\n", GTS_COMMENTS);
-  if (delim == EOF) {
-    g_string_free (buf, TRUE);
-    return 1;
-  }
-  nf = atoi (buf->str);
-  if (delim != '\n')
-    gts_get_newline (fptr);
+  ne = atoi (f->token->str);
 
-  if (nf <= 0) {
-    g_string_free (buf, TRUE);
+  gts_file_next_token (f);
+  if (f->type != GTS_INT) {
+    gts_file_error (f, "expecting an integer (number of faces)");
+    return f->line;
+  }
+  nf = atoi (f->token->str);
+
+  gts_file_first_token_after (f, '\n');
+
+  if (nf <= 0)
     return 0;
-  }
 
   /* allocate nv + 1 just in case nv == 0 */
   vertices = g_malloc ((nv + 1)*sizeof (GtsVertex *));
   edges = g_malloc ((ne + 1)*sizeof (GtsEdge *));
   
   n = 0;
-  while (n < nv && !syntax_error) {
+  while (n < nv && f->type != GTS_ERROR) {
     GtsObject * new_vertex =
       gts_object_new (GTS_OBJECT_CLASS (surface->vertex_class));
 
-    line_number++;
-    delim = (*GTS_OBJECT_CLASS (surface->vertex_class)->read)
-      (&new_vertex, fptr);
-    if (delim == EOF)
-      syntax_error = TRUE;
-    else if (delim != '\n')
-      gts_get_newline (fptr);
-    vertices[n++] = GTS_VERTEX (new_vertex);
+    (* GTS_OBJECT_CLASS (surface->vertex_class)->read) (&new_vertex, f);
+    if (f->type != GTS_ERROR) {
+      gts_file_first_token_after (f, '\n');
+      vertices[n++] = GTS_VERTEX (new_vertex);
+    }
   }
-  if (syntax_error)
+  if (f->type == GTS_ERROR)
     nv = n;
 
   n = 0;
-  while (n < ne && !syntax_error) {
+  while (n < ne && f->type != GTS_ERROR) {
     guint p1, p2;
 
-    line_number++;
-    delim = gts_get_token (fptr, buf, " \t\n", GTS_COMMENTS);
-    if (delim == EOF || delim == '\n')
-      syntax_error = TRUE;
+    if (f->type != GTS_INT)
+      gts_file_error (f, "expecting an integer (first vertex index)");
     else {
-      p1 = atoi (buf->str);
-      delim = gts_get_token (fptr, buf, " \t\n", GTS_COMMENTS);
-      if (delim == EOF)
-	syntax_error = TRUE;
+      p1 = atoi (f->token->str);
+      if (p1 == 0 || p1 > nv)
+	gts_file_error (f, "vertex index `%d' is out of range `[1,%d]'", 
+			p1, nv);
       else {
-	p2 = atoi (buf->str);
-	if (p1 == 0 || p2 == 0 || p1 > nv || p2 > nv)
-	  syntax_error = TRUE;
+	gts_file_next_token (f);
+	if (f->type != GTS_INT)
+	  gts_file_error (f, "expecting an integer (second vertex index)");
 	else {
-	  GtsEdge * new_edge =
-	    gts_edge_new (surface->edge_class,
-			  vertices[p1 - 1], vertices[p2 - 1]);
-	  if (delim != '\n') {
-	    if (GTS_OBJECT_CLASS (surface->edge_class)->read)
-	      delim = (*GTS_OBJECT_CLASS (surface->edge_class)->read)
-		((GtsObject **) &new_edge, fptr);
-	    if (delim != '\n')
-	      gts_get_newline (fptr);
+	  p2 = atoi (f->token->str);
+	  if (p2 == 0 || p2 > nv)
+	    gts_file_error (f, "vertex index `%d' is out of range `[1,%d]'", 
+			    p2, nv);
+	  else {
+	    GtsEdge * new_edge =
+	      gts_edge_new (surface->edge_class,
+			    vertices[p1 - 1], vertices[p2 - 1]);
+
+	    gts_file_next_token (f);
+	    if (f->type != '\n')
+	      if (GTS_OBJECT_CLASS (surface->edge_class)->read)
+		(*GTS_OBJECT_CLASS (surface->edge_class)->read)
+		  ((GtsObject **) &new_edge, f);
+	    gts_file_first_token_after (f, '\n');
+	    edges[n++] = new_edge;
 	  }
-	  edges[n++] = new_edge;
 	}
       }
     }
   }
-  if (syntax_error)
+  if (f->type == GTS_ERROR)
     ne = n;
 
   n = 0;
-  while (n < nf && !syntax_error) {
+  while (n < nf && f->type != GTS_ERROR) {
     guint s1, s2, s3;
 
-    line_number++;
-    delim = gts_get_token (fptr, buf, " \t\n", GTS_COMMENTS);
-    if (delim == EOF || delim == '\n')
-      syntax_error = TRUE;
+    if (f->type != GTS_INT)
+      gts_file_error (f, "expecting an integer (first edge index)");
     else {
-      s1 = atoi (buf->str);
-      delim = gts_get_token (fptr, buf, " \t\n", GTS_COMMENTS);
-      if (delim == EOF || delim == '\n')
-	syntax_error = TRUE;
+      s1 = atoi (f->token->str);
+      if (s1 == 0 || s1 > ne)
+	gts_file_error (f, "edge index `%d' is out of range `[1,%d]'", 
+			s1, ne);
       else {
-	s2 = atoi (buf->str);
-	delim = gts_get_token (fptr, buf, " \t\n", GTS_COMMENTS);
-	if (delim == EOF)
-	  syntax_error = TRUE;
+	gts_file_next_token (f);
+	if (f->type != GTS_INT)
+	  gts_file_error (f, "expecting an integer (second edge index)");
 	else {
-	  s3 = atoi (buf->str);
-	  if (s1 == 0 || s2 == 0 || s3 == 0 || s1 > ne || s2 > ne || s3 > ne)
-	    syntax_error = TRUE;
+	  s2 = atoi (f->token->str);
+	  if (s2 == 0 || s2 > ne)
+	    gts_file_error (f, "edge index `%d' is out of range `[1,%d]'", 
+			    s2, ne);
 	  else {
-	    GtsFace * new_face = gts_face_new (surface->face_class,
-					       edges[s1 - 1],
-					       edges[s2 - 1],
-					       edges[s3 - 1]);
-	    if (delim != '\n') {
-	      if (GTS_OBJECT_CLASS (surface->face_class)->read)
-		delim = (*GTS_OBJECT_CLASS (surface->face_class)->read)
-		  ((GtsObject **) &new_face, fptr);
-	      if (delim != '\n')
-		gts_get_newline (fptr);
+	    gts_file_next_token (f);
+	    if (f->type != GTS_INT)
+	      gts_file_error (f, "expecting an integer (third edge index)");
+	    else {
+	      s3 = atoi (f->token->str);
+	      if (s3 == 0 || s3 > ne)
+		gts_file_error (f, "edge index `%d' is out of range `[1,%d]'", 
+				s3, ne);
+	      else {
+		GtsFace * new_face = gts_face_new (surface->face_class,
+						   edges[s1 - 1],
+						   edges[s2 - 1],
+						   edges[s3 - 1]);
+
+		gts_file_next_token (f);
+		if (f->type != '\n')
+		  if (GTS_OBJECT_CLASS (surface->face_class)->read)
+		    (*GTS_OBJECT_CLASS (surface->face_class)->read)
+		      ((GtsObject **) &new_face, f);
+		gts_file_first_token_after (f, '\n');
+		gts_surface_add_face (surface, new_face);
+		n++;
+	      }
 	    }
-	    gts_surface_add_face (surface, new_face);
-	    n++;
 	  }
 	}
       }
     }
   }
 
-  if (syntax_error) {
+  if (f->type == GTS_ERROR) {
     gts_allow_floating_vertices = TRUE;
     while (nv)
       gts_object_destroy (GTS_OBJECT (vertices[nv-- - 1]));
     gts_allow_floating_vertices = FALSE;
-    gts_object_destroy (GTS_OBJECT (surface));
   }
 
   g_free (vertices);
   g_free (edges);
-  g_string_free (buf, TRUE);
 
-  if (syntax_error)
-    return line_number;
+  if (f->type == GTS_ERROR)
+    return f->line;
   return 0;
 }
 
@@ -361,7 +398,7 @@ void gts_range_init (GtsRange * r)
 
   r->max = - G_MAXDOUBLE;
   r->min = G_MAXDOUBLE;
-  r->mean = r->stddev = 0.0;
+  r->sum = r->sum2 = 0.0;
   r->n = 0;
 }
 
@@ -377,7 +414,7 @@ void gts_range_reset (GtsRange * r)
 
   r->max = 0.0;
   r->min = 0.0;
-  r->mean = r->stddev = 0.0;
+  r->sum = r->sum2 = 0.0;
   r->n = 0;
 }
 
@@ -394,8 +431,8 @@ void gts_range_add_value (GtsRange * r, gdouble val)
 
   if (val < r->min) r->min = val;
   if (val > r->max) r->max = val;
-  r->mean += val;
-  r->stddev += val*val;
+  r->sum += val;
+  r->sum2 += val*val;
   r->n++;
 }
 
@@ -408,16 +445,17 @@ void gts_range_add_value (GtsRange * r, gdouble val)
 void gts_range_update (GtsRange * r)
 {
   g_return_if_fail (r != NULL);
+
   if (r->n > 0) {
-    if (r->stddev - r->mean*r->mean/(gdouble)r->n >= 0.)
-      r->stddev = sqrt ((r->stddev - r->mean*r->mean/(gdouble)r->n)
-			/(gdouble)r->n);
+    if (r->sum2 - r->sum*r->sum/(gdouble) r->n >= 0.)
+      r->stddev = sqrt ((r->sum2 - r->sum*r->sum/(gdouble) r->n)
+			/(gdouble) r->n);
     else
       r->stddev = 0.;
-    r->mean /= (gdouble)r->n;
+    r->mean = r->sum/(gdouble) r->n;
   }
-  else
-    r->min = r->max = 0.0;
+  else 
+    r->min = r->max = r->mean = r->stddev = 0.;
 }
 
 /**
@@ -635,44 +673,31 @@ static void write_vertex (GtsPoint * p, gpointer * data)
 {
   (*GTS_OBJECT (p)->klass->write) (GTS_OBJECT (p), (FILE *) data[0]);
   fputc ('\n', (FILE *) data[0]);
-  GTS_OBJECT (p)->reserved = GUINT_TO_POINTER (++(*((guint *) data[1])));
+  g_hash_table_insert (data[2], p, 
+		       GUINT_TO_POINTER (++(*((guint *) data[1]))));
 }
 
 static void write_edge (GtsSegment * s, gpointer * data) 
 {
   fprintf ((FILE *) data[0], "%u %u",
-	   GPOINTER_TO_UINT (GTS_OBJECT (s->v1)->reserved),
-	   GPOINTER_TO_UINT (GTS_OBJECT (s->v2)->reserved));
+	   GPOINTER_TO_UINT (g_hash_table_lookup (data[2], s->v1)),
+	   GPOINTER_TO_UINT (g_hash_table_lookup (data[2], s->v2)));
   if (GTS_OBJECT (s)->klass->write)
     (*GTS_OBJECT (s)->klass->write) (GTS_OBJECT (s), (FILE *) data[0]);
   fputc ('\n', (FILE *) data[0]);
-  GTS_OBJECT (s)->reserved = GUINT_TO_POINTER (++(*((guint *) data[1])));
+  g_hash_table_insert (data[3], s, 
+		       GUINT_TO_POINTER (++(*((guint *) data[1]))));
 }
 
-static void write_face (GtsTriangle * t, FILE * fp)
+static void write_face (GtsTriangle * t, gpointer * data)
 {
-  fprintf (fp, "%u %u %u",
-	   GPOINTER_TO_UINT (GTS_OBJECT (t->e1)->reserved),
-	   GPOINTER_TO_UINT (GTS_OBJECT (t->e2)->reserved),
-	   GPOINTER_TO_UINT (GTS_OBJECT (t->e3)->reserved));
+  fprintf (data[0], "%u %u %u",
+	   GPOINTER_TO_UINT (g_hash_table_lookup (data[3], t->e1)),
+	   GPOINTER_TO_UINT (g_hash_table_lookup (data[3], t->e2)),
+	   GPOINTER_TO_UINT (g_hash_table_lookup (data[3], t->e3)));
   if (GTS_OBJECT (t)->klass->write)
-    (*GTS_OBJECT (t)->klass->write) (GTS_OBJECT (t), fp);
-  fputc ('\n', fp);
-}
-
-static void reset_reserved (GtsTriangle * t)
-{
-  GtsSegment
-    * s1 = GTS_SEGMENT (t->e1), 
-    * s2 = GTS_SEGMENT (t->e2), 
-    * s3 = GTS_SEGMENT (t->e3);
-
-  GTS_OBJECT (s1)->reserved = NULL;
-  GTS_OBJECT (s2)->reserved = NULL;
-  GTS_OBJECT (s3)->reserved = NULL;
-  GTS_OBJECT (s1->v1)->reserved = GTS_OBJECT (s1->v2)->reserved = NULL;
-  GTS_OBJECT (s2->v1)->reserved = GTS_OBJECT (s2->v2)->reserved = NULL;
-  GTS_OBJECT (s3->v1)->reserved = GTS_OBJECT (s3->v2)->reserved = NULL;
+    (*GTS_OBJECT (t)->klass->write) (GTS_OBJECT (t), data[0]);
+  fputc ('\n', data[0]);
 }
 
 /**
@@ -705,7 +730,8 @@ static void reset_reserved (GtsTriangle * t)
 void gts_surface_write (GtsSurface * s, FILE * fptr)
 {
   guint n;
-  gpointer data[2];
+  gpointer data[4];
+  GHashTable * vindex, * eindex;
   GtsSurfaceStats stats;
 
   g_return_if_fail (s != NULL);
@@ -713,6 +739,8 @@ void gts_surface_write (GtsSurface * s, FILE * fptr)
 
   data[0] = fptr;
   data[1] = &n;
+  data[2] = vindex = g_hash_table_new (NULL, NULL);
+  data[3] = eindex = g_hash_table_new (NULL, NULL);
 
   gts_surface_stats (s, &stats);
   fprintf (fptr, "%u %u %u", 
@@ -726,8 +754,9 @@ void gts_surface_write (GtsSurface * s, FILE * fptr)
   gts_surface_foreach_vertex (s, (GtsFunc) write_vertex, data);
   n = 0;
   gts_surface_foreach_edge (s, (GtsFunc) write_edge, data);
-  gts_surface_foreach_face (s, (GtsFunc) write_face, fptr);
-  gts_surface_foreach_face (s, (GtsFunc) reset_reserved, NULL);
+  gts_surface_foreach_face (s, (GtsFunc) write_face, data);
+  g_hash_table_destroy (vindex);
+  g_hash_table_destroy (eindex);
 }
 
 static void write_vertex_oogl (GtsPoint * p, gpointer * data)
@@ -833,13 +862,19 @@ void gts_surface_write_oogl_boundary (GtsSurface * s, FILE * fptr)
   fputs ("}\n", fptr);
 }
 
+#ifdef USE_SURFACE_BTREE
+static gint vertex_foreach_face (GtsTriangle * t,
+				 gpointer t_data,
+				 gpointer * info)
+#else /* not USE_SURFACE_BTREE */
 static void vertex_foreach_face (GtsTriangle * t,
 				 gpointer t_data,
 				 gpointer * info)
+#endif /* not USE_SURFACE_BTREE */
 {
   GHashTable * hash = info[0];
   gpointer data = info[1];
-  GtsFunc func = info[2];
+  GtsFunc func = (GtsFunc) info[2];
   GtsSegment 
     * s1 = GTS_SEGMENT (t->e1);
 
@@ -856,6 +891,9 @@ static void vertex_foreach_face (GtsTriangle * t,
     g_hash_table_insert (hash, gts_triangle_vertex (t), 
 			 GINT_TO_POINTER (-1));
   }
+#ifdef USE_SURFACE_BTREE
+  return FALSE;
+#endif /* USE_SURFACE_BTREE */
 }
 
 /**
@@ -878,19 +916,30 @@ void gts_surface_foreach_vertex (GtsSurface * s, GtsFunc func, gpointer data)
   info[0] = g_hash_table_new (NULL, NULL);
   info[1] = data;
   info[2] = func;
+#ifdef USE_SURFACE_BTREE
+  g_tree_traverse (s->faces, (GTraverseFunc) vertex_foreach_face, G_IN_ORDER,
+		   info);
+#else /* not USE_SURFACE_BTREE */
   g_hash_table_foreach (s->faces, (GHFunc) vertex_foreach_face, info);
+#endif /* not USE_SURFACE_BTREE */
   g_hash_table_destroy (info[0]);
   /* allow removal of faces */
   s->keep_faces = FALSE;
 }
 
+#ifdef USE_SURFACE_BTREE
+static gint edge_foreach_face (GtsTriangle * t,
+			       gpointer t_data, 
+			       gpointer * info)
+#else /* not USE_SURFACE_BTREE */
 static void edge_foreach_face (GtsTriangle * t,
 			       gpointer t_data, 
 			       gpointer * info)
+#endif /* not USE_SURFACE_BTREE */
 {
   GHashTable * hash = info[0];
   gpointer data = info[1];
-  GtsFunc func = info[2];
+  GtsFunc func = (GtsFunc) info[2];
 
   if (!g_hash_table_lookup (hash, t->e1)) {
     (*func) (t->e1, data);
@@ -904,6 +953,9 @@ static void edge_foreach_face (GtsTriangle * t,
     (*func) (t->e3, data);
     g_hash_table_insert (hash, t->e3, GINT_TO_POINTER (-1));
   }
+#ifdef USE_SURFACE_BTREE
+  return FALSE;
+#endif /* not USE_SURFACE_BTREE */
 }
 
 /**
@@ -926,71 +978,31 @@ void gts_surface_foreach_edge (GtsSurface * s, GtsFunc func, gpointer data)
   info[0] = g_hash_table_new (NULL, NULL);
   info[1] = data;
   info[2] = func;
-  g_hash_table_foreach (s->faces, (GHFunc)edge_foreach_face, info);
+#ifdef USE_SURFACE_BTREE
+  g_tree_traverse (s->faces, (GTraverseFunc) edge_foreach_face, G_IN_ORDER,
+		   info);
+#else /* not USE_SURFACE_BTREE */
+  g_hash_table_foreach (s->faces, (GHFunc) edge_foreach_face, info);
+#endif /* not USE_SURFACE_BTREE */
   g_hash_table_destroy (info[0]);
   /* allow removal of faces */
   s->keep_faces = FALSE;
 }
 
-#if 0
-
-/* Due to an unkown reason g_hash_table_foreach does not allow to stop 
- * the loop, hence the redefinition. I hope they don't change
- * the GHashTable, GHashNode structures ... */
-typedef struct _GHashNode      GHashNode;
-
-struct _GHashNode
-{
-  gpointer key;
-  gpointer value;
-  GHashNode *next;
-};
-
-struct _GHashTable
-{
-  gint size;
-  gint nnodes;
-  guint frozen;
-  GHashNode **nodes;
-  GHashFunc hash_func;
-  GCompareFunc key_compare_func;
-};
-
-/**
- * gts_surface_foreach_face:
- * @s: a #GtsSurface.
- * @func: a #GtsFunc.
- * @data: user data to be passed to @func.
- *
- * Calls @func once for each face of @s.
- */
-void gts_surface_foreach_face (GtsSurface * s,
-			       GtsFunc func, 
-			       gpointer data)
-{
-  guint i;
-  GHashNode * node;
-  gboolean stop = FALSE;
-
-  g_return_if_fail (s != NULL);
-  g_return_if_fail (func != NULL);
-
-  /* forbid removal of faces */
-  s->keep_faces = TRUE;
-  for (i = 0; i < s->faces->size && !stop; i++)
-    for (node = s->faces->nodes[i]; node && !stop; node = node->next)
-      stop = !(* func) (node->key, data);
-  /* allow removal of faces */
-  s->keep_faces = FALSE;
-}
-
-#else
-
+#ifdef USE_SURFACE_BTREE
+static gint foreach_face (GtsFace * f, 
+			  gpointer t_data,
+			  gpointer * info)
+#else /* not USE_SURFACE_BTREE */
 static void foreach_face (GtsFace * f, 
 			  gpointer t_data,
 			  gpointer * info)
+#endif /* not USE_SURFACE_BTREE */
 {
   (*((GtsFunc) info[0])) (f, info[1]);
+#ifdef USE_SURFACE_BTREE
+  return FALSE;
+#endif /* USE_SURFACE_BTREE */
 }
 
 /**
@@ -1014,13 +1026,40 @@ void gts_surface_foreach_face (GtsSurface * s,
   s->keep_faces = TRUE;
   info[0] = func;
   info[1] = data;
-  g_hash_table_foreach (s->faces, (GHFunc)foreach_face, info);
+#ifdef USE_SURFACE_BTREE
+  g_tree_traverse (s->faces, (GTraverseFunc) foreach_face, G_IN_ORDER,
+		   info);
+#else /* not USE_SURFACE_BTREE */
+  g_hash_table_foreach (s->faces, (GHFunc) foreach_face, info);
+#endif /* not USE_SURFACE_BTREE */
   /* allow removal of faces */
   s->keep_faces = FALSE;
 }
 
-#endif
+#ifdef USE_SURFACE_BTREE
+static gint foreach_face_remove (GtsFace * f,
+				 gpointer t_data,
+				 gpointer * info)
+{
+  if ((*((GtsFunc) info[0])) (f, info[1])) {
+    GtsSurface * s = info[2];
+    guint * n = info[3];
 
+    f->surfaces = g_slist_remove (f->surfaces, s);
+    if (!GTS_OBJECT_DESTROYED (f) &&
+	!gts_allow_floating_faces && 
+	f->surfaces == NULL)
+      gts_object_destroy (GTS_OBJECT (f));
+    
+    if (GTS_SURFACE_CLASS (GTS_OBJECT (s)->klass)->remove_face)
+      (* GTS_SURFACE_CLASS (GTS_OBJECT (s)->klass)->remove_face) (s, f);
+
+    g_tree_remove (s->faces, f);
+    (*n)++;
+  }
+  return FALSE;
+}
+#else /* not USE_SURFACE_BTREE */
 static gboolean foreach_face_remove (GtsFace * f,
 				     gpointer t_data,
 				     gpointer * info)
@@ -1041,6 +1080,7 @@ static gboolean foreach_face_remove (GtsFace * f,
   }
   return FALSE;
 }
+#endif /* not USE_SURFACE_BTREE */
 
 /**
  * gts_surface_foreach_face_remove:
@@ -1059,8 +1099,8 @@ guint gts_surface_foreach_face_remove (GtsSurface * s,
 				       GtsFunc func, 
 				       gpointer data)
 {
-  gpointer info[3];
-  guint n;
+  gpointer info[4];
+  guint n = 0;
 
   g_return_val_if_fail (s != NULL, 0);
   g_return_val_if_fail (func != NULL, 0);
@@ -1070,9 +1110,15 @@ guint gts_surface_foreach_face_remove (GtsSurface * s,
   info[0] = func;
   info[1] = data;
   info[2] = s;
+#ifdef USE_SURFACE_BTREE
+  info[3] = &n;
+  g_tree_traverse (s->faces, (GTraverseFunc) foreach_face_remove, G_PRE_ORDER,
+		   info);
+#else /* not USE_SURFACE_BTREE */
   n = g_hash_table_foreach_remove (s->faces, 
 				   (GHRFunc) foreach_face_remove, 
 				   info);
+#endif /* not USE_SURFACE_BTREE */
   /* allow removal of faces */
   s->keep_faces = FALSE;
   
@@ -1081,7 +1127,9 @@ guint gts_surface_foreach_face_remove (GtsSurface * s,
 
 static void midvertex_insertion (GtsEdge * e,
 				 GtsSurface * surface,
-				 GtsHeap * heap,
+				 GtsEHeap * heap,
+				 GtsRefineFunc refine_func,
+				 gpointer refine_data,
 				 GtsVertexClass * vertex_class,
 				 GtsEdgeClass * edge_class)
 {
@@ -1089,11 +1137,11 @@ static void midvertex_insertion (GtsEdge * e,
   GtsEdge * e1, * e2;
   GSList * i;
 
-  midvertex = gts_segment_midvertex (GTS_SEGMENT (e), vertex_class);
+  midvertex = (*refine_func) (e, vertex_class, refine_data);
   e1 = gts_edge_new (edge_class, GTS_SEGMENT (e)->v1, midvertex);
-  gts_heap_insert (heap, e1);
+  gts_eheap_insert (heap, e1);
   e2 = gts_edge_new (edge_class, GTS_SEGMENT (e)->v2, midvertex);
-  gts_heap_insert (heap, e2);
+  gts_eheap_insert (heap, e2);
   
   /* creates new faces and modifies old ones */
   i = e->triangles;
@@ -1104,7 +1152,7 @@ static void midvertex_insertion (GtsEdge * e,
 
     gts_triangle_vertices_edges (t, e, &v1, &v2, &v3, &e, &te2, &te3);
     ne = gts_edge_new (edge_class, midvertex, v3);
-    gts_heap_insert (heap, ne);
+    gts_eheap_insert (heap, ne);
     if (GTS_SEGMENT (e1)->v1 == v2) {
       tmp = e1; e1 = e2; e2 = tmp;
     }
@@ -1122,61 +1170,69 @@ static void midvertex_insertion (GtsEdge * e,
   gts_object_destroy (GTS_OBJECT (e));
 }
 
-static gint edge_is_longer (GtsSegment * s1, GtsSegment * s2)
+static gdouble edge_length2_inverse (GtsSegment * s)
 {
-  if (gts_point_distance2 (GTS_POINT (s1->v1), GTS_POINT (s1->v2)) >
-      gts_point_distance2 (GTS_POINT (s2->v1), GTS_POINT (s2->v2)))
-    return -1;
-  return 1;
+  return - gts_point_distance2 (GTS_POINT (s->v1), GTS_POINT (s->v2));
 }
 
-static void create_heap_refine (GtsEdge * e, GtsHeap * heap)
+static void create_heap_refine (GtsEdge * e, GtsEHeap * heap)
 {
-  gts_heap_insert (heap, e);
+  gts_eheap_insert (heap, e);
 }
 
 /**
  * gts_surface_refine:
  * @surface: a #GtsSurface.
- * @compare_func: a GCompareFunc.
+ * @cost_func: a function returning the cost for a given edge.
+ * @cost_data: user data to be passed to @cost_func.
+ * @refine_func: a #GtsRefineFunc.
+ * @refine_data: user data to be passed to @refine_func.
  * @stop_func: a #GtsStopFunc.
- * @data: user data to be passed to @stop_func.
+ * @stop_data: user data to be passed to @stop_func.
  *
  * Refine @surface using a midvertex insertion technique. All the
- * edges of @surface are ordered according to @compare_func. The edges
+ * edges of @surface are ordered according to @cost_func. The edges
  * are then processed in order until @stop_func returns %TRUE. Each
  * edge is split in two and new edges and faces are created.
  *
- * If @compare_func is set to %NULL an internal function comparing the
- * squared length of the edges is used.  
+ * If @cost_func is set to %NULL, the edges are sorted according 
+ * to their length squared (the longest is on top).
+ *
+ * If @refine_func is set to %NULL gts_segment_midvertex() is used.
+ * 
  */
 void gts_surface_refine (GtsSurface * surface,
-			 GCompareFunc compare_func,
+			 GtsKeyFunc cost_func,
+			 gpointer cost_data,
+			 GtsRefineFunc refine_func,
+			 gpointer refine_data,
 			 GtsStopFunc stop_func,
-			 gpointer data)
+			 gpointer stop_data)
 {
-  GtsHeap * heap;
+  GtsEHeap * heap;
   GtsEdge * e;
+  gdouble top_cost;
 
   g_return_if_fail (surface != NULL);
   g_return_if_fail (stop_func != NULL);
 
-  if (compare_func == NULL)
-    compare_func = (GCompareFunc) edge_is_longer;
+  if (cost_func == NULL)
+    cost_func = (GtsKeyFunc) edge_length2_inverse;
+  if (refine_func == NULL)
+    refine_func = (GtsRefineFunc) gts_segment_midvertex;
 
-  heap = gts_heap_new (compare_func);
-  gts_heap_freeze (heap);
-  gts_surface_foreach_edge (surface, (GtsFunc)create_heap_refine, heap);
-  gts_heap_thaw (heap);
-  while ((e = gts_heap_remove_top (heap)) &&
-	 !(*stop_func) (gts_point_distance (GTS_POINT (GTS_SEGMENT (e)->v1), 
-					    GTS_POINT (GTS_SEGMENT (e)->v2)),
-			gts_heap_size (heap) + 
+  heap = gts_eheap_new (cost_func, cost_data);
+  gts_eheap_freeze (heap);
+  gts_surface_foreach_edge (surface, (GtsFunc) create_heap_refine, heap);
+  gts_eheap_thaw (heap);
+  while ((e = gts_eheap_remove_top (heap, &top_cost)) &&
+	 !(*stop_func) (top_cost,
+			gts_eheap_size (heap) + 
 			gts_edge_face_number (e, surface) + 2,
-			data))
-    midvertex_insertion (e, surface, heap, 
+			stop_data))
+    midvertex_insertion (e, surface, heap, refine_func, refine_data,
 			 surface->vertex_class, surface->edge_class);
-  gts_heap_destroy (heap);
+  gts_eheap_destroy (heap);
 }
 
 static GSList * edge_triangles (GtsEdge * e1, GtsEdge * e)
@@ -1614,6 +1670,315 @@ gboolean gts_coarsen_stop_cost (gdouble cost,
   return FALSE;
 }
 
+#define GTS_M_ICOSAHEDRON_X /* sqrt(sqrt(5)+1)/sqrt(2*sqrt(5)) */ \
+  0.850650808352039932181540497063011072240401406
+#define GTS_M_ICOSAHEDRON_Y /* sqrt(2)/sqrt(5+sqrt(5))         */ \
+  0.525731112119133606025669084847876607285497935
+#define GTS_M_ICOSAHEDRON_Z 0.0
+
+static guint generate_icosahedron (GtsSurface * s)
+{
+  GtsVertex * v01 = gts_vertex_new (s->vertex_class,
+      +GTS_M_ICOSAHEDRON_Z, +GTS_M_ICOSAHEDRON_X, -GTS_M_ICOSAHEDRON_Y);
+  GtsVertex * v02 = gts_vertex_new (s->vertex_class,
+      +GTS_M_ICOSAHEDRON_X, +GTS_M_ICOSAHEDRON_Y, +GTS_M_ICOSAHEDRON_Z);
+  GtsVertex * v03 = gts_vertex_new (s->vertex_class,
+      +GTS_M_ICOSAHEDRON_Y, +GTS_M_ICOSAHEDRON_Z, -GTS_M_ICOSAHEDRON_X);
+  GtsVertex * v04 = gts_vertex_new (s->vertex_class,
+      +GTS_M_ICOSAHEDRON_Y, +GTS_M_ICOSAHEDRON_Z, +GTS_M_ICOSAHEDRON_X);
+  GtsVertex * v05 = gts_vertex_new (s->vertex_class,
+      +GTS_M_ICOSAHEDRON_X, -GTS_M_ICOSAHEDRON_Y, +GTS_M_ICOSAHEDRON_Z);
+  GtsVertex * v06 = gts_vertex_new (s->vertex_class,
+      +GTS_M_ICOSAHEDRON_Z, +GTS_M_ICOSAHEDRON_X, +GTS_M_ICOSAHEDRON_Y);
+  GtsVertex * v07 = gts_vertex_new (s->vertex_class,
+      -GTS_M_ICOSAHEDRON_Y, +GTS_M_ICOSAHEDRON_Z, +GTS_M_ICOSAHEDRON_X);
+  GtsVertex * v08 = gts_vertex_new (s->vertex_class,
+      +GTS_M_ICOSAHEDRON_Z, -GTS_M_ICOSAHEDRON_X, -GTS_M_ICOSAHEDRON_Y);
+  GtsVertex * v09 = gts_vertex_new (s->vertex_class,
+      -GTS_M_ICOSAHEDRON_X, +GTS_M_ICOSAHEDRON_Y, +GTS_M_ICOSAHEDRON_Z);
+  GtsVertex * v10 = gts_vertex_new (s->vertex_class,
+      -GTS_M_ICOSAHEDRON_Y, +GTS_M_ICOSAHEDRON_Z, -GTS_M_ICOSAHEDRON_X);
+  GtsVertex * v11 = gts_vertex_new (s->vertex_class,
+      -GTS_M_ICOSAHEDRON_X, -GTS_M_ICOSAHEDRON_Y, +GTS_M_ICOSAHEDRON_Z);
+  GtsVertex * v12 = gts_vertex_new (s->vertex_class,
+      +GTS_M_ICOSAHEDRON_Z, -GTS_M_ICOSAHEDRON_X, +GTS_M_ICOSAHEDRON_Y);
+
+  GtsEdge * e01 = gts_edge_new (s->edge_class, v01, v02);
+  GtsEdge * e02 = gts_edge_new (s->edge_class, v03, v02);
+  GtsEdge * e03 = gts_edge_new (s->edge_class, v01, v03);
+  GtsEdge * e04 = gts_edge_new (s->edge_class, v04, v05);
+  GtsEdge * e05 = gts_edge_new (s->edge_class, v02, v05);
+  GtsEdge * e06 = gts_edge_new (s->edge_class, v04, v02);
+  GtsEdge * e07 = gts_edge_new (s->edge_class, v06, v07);
+  GtsEdge * e08 = gts_edge_new (s->edge_class, v04, v07);
+  GtsEdge * e09 = gts_edge_new (s->edge_class, v06, v04);
+  GtsEdge * e10 = gts_edge_new (s->edge_class, v08, v03);
+  GtsEdge * e11 = gts_edge_new (s->edge_class, v03, v05);
+  GtsEdge * e12 = gts_edge_new (s->edge_class, v08, v05);
+  GtsEdge * e13 = gts_edge_new (s->edge_class, v06, v09);
+  GtsEdge * e14 = gts_edge_new (s->edge_class, v07, v09);
+  GtsEdge * e15 = gts_edge_new (s->edge_class, v08, v10);
+  GtsEdge * e16 = gts_edge_new (s->edge_class, v03, v10);
+  GtsEdge * e17 = gts_edge_new (s->edge_class, v06, v01);
+  GtsEdge * e18 = gts_edge_new (s->edge_class, v01, v09);
+  GtsEdge * e19 = gts_edge_new (s->edge_class, v08, v11);
+  GtsEdge * e20 = gts_edge_new (s->edge_class, v10, v11);
+  GtsEdge * e21 = gts_edge_new (s->edge_class, v06, v02);
+  GtsEdge * e22 = gts_edge_new (s->edge_class, v12, v11);
+  GtsEdge * e23 = gts_edge_new (s->edge_class, v12, v08);
+  GtsEdge * e24 = gts_edge_new (s->edge_class, v12, v07);
+  GtsEdge * e25 = gts_edge_new (s->edge_class, v07, v11);
+  GtsEdge * e26 = gts_edge_new (s->edge_class, v12, v04);
+  GtsEdge * e27 = gts_edge_new (s->edge_class, v09, v11);
+  GtsEdge * e28 = gts_edge_new (s->edge_class, v10, v09);
+  GtsEdge * e29 = gts_edge_new (s->edge_class, v12, v05);
+  GtsEdge * e30 = gts_edge_new (s->edge_class, v01, v10);
+  
+  gts_surface_add_face (s, gts_face_new (s->face_class, e01, e02, e03));
+  gts_surface_add_face (s, gts_face_new (s->face_class, e04, e05, e06));
+  gts_surface_add_face (s, gts_face_new (s->face_class, e07, e08, e09));
+  gts_surface_add_face (s, gts_face_new (s->face_class, e10, e11, e12));
+  gts_surface_add_face (s, gts_face_new (s->face_class, e13, e14, e07));
+  gts_surface_add_face (s, gts_face_new (s->face_class, e15, e16, e10));
+  gts_surface_add_face (s, gts_face_new (s->face_class, e17, e18, e13));
+  gts_surface_add_face (s, gts_face_new (s->face_class, e19, e20, e15));
+  gts_surface_add_face (s, gts_face_new (s->face_class, e21, e01, e17));
+  gts_surface_add_face (s, gts_face_new (s->face_class, e22, e19, e23));
+  gts_surface_add_face (s, gts_face_new (s->face_class, e09, e06, e21));
+  gts_surface_add_face (s, gts_face_new (s->face_class, e24, e25, e22));
+  gts_surface_add_face (s, gts_face_new (s->face_class, e26, e08, e24));
+  gts_surface_add_face (s, gts_face_new (s->face_class, e20, e27, e28));
+  gts_surface_add_face (s, gts_face_new (s->face_class, e29, e04, e26));
+  gts_surface_add_face (s, gts_face_new (s->face_class, e14, e27, e25));
+  gts_surface_add_face (s, gts_face_new (s->face_class, e23, e12, e29));
+  gts_surface_add_face (s, gts_face_new (s->face_class, e02, e05, e11));
+  gts_surface_add_face (s, gts_face_new (s->face_class, e30, e28, e18));
+  gts_surface_add_face (s, gts_face_new (s->face_class, e03, e16, e30));
+
+  return 0;
+}
+
+static GtsVertex * unit_sphere_arc_midvertex (GtsSegment * s, 
+					      GtsVertexClass * vertex_class)
+{
+  GtsPoint * p1, * p2;
+  gdouble x, y, z, norm;
+
+  p1 = GTS_POINT (s->v1); p2 = GTS_POINT (s->v2);
+
+  x = 0.5*(p1->x + p2->x);
+  y = 0.5*(p1->y + p2->y);
+  z = 0.5*(p1->z + p2->z);
+
+  norm = x*x + y*y + z*z;
+  norm = sqrt (norm);
+
+  x /= norm; y /= norm; z /= norm;
+
+  return gts_vertex_new (vertex_class, x, y, z);
+}
+
+static void tessellate_face (GtsFace * f,
+			     GtsSurface * s,
+			     GtsRefineFunc refine_func,
+			     gpointer refine_data,
+			     GtsVertexClass * vertex_class,
+			     GtsEdgeClass * edge_class)
+{
+  GtsTriangle * t;
+  GtsEdge * e1, * e2, * e3;                          /* former edges     */
+  GtsVertex * v1, * v2, * v3;                        /* initial vertices */
+  GtsVertex * v4, * v5, * v6;                        /* new vertices     */ 
+  GtsEdge * e56, * e64, * e45;                       /* new inside edges */
+  GtsEdge * e24, * e34, * e35, * e15, * e16, * e26;  /* new border edges */
+  GSList * dum;
+  GtsEdge * edum;
+  
+  t = GTS_TRIANGLE (f);
+  e1 = t->e1; e2 = t->e2; e3 = t->e3;
+
+  if (GTS_SEGMENT (e1)->v2 == GTS_SEGMENT (e2)->v1) {
+    v1 = GTS_SEGMENT (e2)->v2;
+    v2 = GTS_SEGMENT (e1)->v1;
+    v3 = GTS_SEGMENT (e1)->v2;
+  }
+  else if (GTS_SEGMENT (e1)->v2 == GTS_SEGMENT (e2)->v2) {
+    v1 = GTS_SEGMENT (e2)->v1;
+    v2 = GTS_SEGMENT (e1)->v1;
+    v3 = GTS_SEGMENT (e1)->v2;
+  }
+  else if (GTS_SEGMENT (e1)->v1 == GTS_SEGMENT (e2)->v1) {
+    v1 = GTS_SEGMENT (e2)->v2;
+    v2 = GTS_SEGMENT (e1)->v2;
+    v3 = GTS_SEGMENT (e1)->v1;
+  }
+  else if (GTS_SEGMENT (e1)->v1 == GTS_SEGMENT (e2)->v2) {
+    v1 = GTS_SEGMENT (e2)->v1;
+    v2 = GTS_SEGMENT (e1)->v2;
+    v3 = GTS_SEGMENT (e1)->v1;
+  }
+  else {
+    v1 = v2 = v3 = NULL;
+    g_assert_not_reached ();
+  }
+
+  e1->triangles = g_slist_remove (e1->triangles, t);
+  e2->triangles = g_slist_remove (e2->triangles, t);
+  e3->triangles = g_slist_remove (e3->triangles, t);
+  
+  if (GTS_OBJECT (e1)->reserved) {
+    dum = (GTS_OBJECT (e1)->reserved);
+    e24 = dum->data;
+    e34 = dum->next->data;
+    v4 = GTS_SEGMENT (e24)->v2;
+    if (GTS_SEGMENT (e24)->v1 == v3) {
+      edum = e34; e34 = e24; e24 = edum;
+    }
+  }
+  else {
+    v4 = (*refine_func) (e1, vertex_class, refine_data);
+    e24 = gts_edge_new (edge_class, v2, v4);
+    e34 = gts_edge_new (edge_class, v3, v4);
+    dum = g_slist_append (NULL, e24);
+    dum = g_slist_append (dum,  e34);
+    GTS_OBJECT (e1)->reserved = dum;
+  }
+  if (GTS_OBJECT (e2)->reserved) {
+    dum = (GTS_OBJECT (e2)->reserved);
+    e35 = dum->data;
+    e15 = dum->next->data;
+    v5 = GTS_SEGMENT (e35)->v2;
+    if (GTS_SEGMENT (e35)->v1 == v1) {
+      edum = e15; e15 = e35; e35 = edum;
+    }
+  }
+  else {
+    v5 = (*refine_func) (e2, vertex_class, refine_data);
+    e35 = gts_edge_new (edge_class, v3, v5);
+    e15 = gts_edge_new (edge_class, v1, v5);
+    dum = g_slist_append (NULL, e35);
+    dum = g_slist_append (dum,  e15);
+    GTS_OBJECT (e2)->reserved = dum;
+  }
+  if (GTS_OBJECT (e3)->reserved) {
+    dum = (GTS_OBJECT (e3)->reserved);
+    e16 = dum->data;
+    e26 = dum->next->data;
+    v6 = GTS_SEGMENT (e16)->v2;
+    if (GTS_SEGMENT (e16)->v1 == v2) {
+      edum = e16; e16 = e26; e26 = edum;
+    }
+  }
+  else {
+    v6 = (*refine_func) (e3, vertex_class, refine_data);
+    e16 = gts_edge_new (edge_class, v1, v6);
+    e26 = gts_edge_new (edge_class, v2, v6);
+    dum = g_slist_append (NULL, e16);
+    dum = g_slist_append (dum,  e26);
+    GTS_OBJECT (e3)->reserved = dum;
+  }
+  
+  if (e1->triangles == NULL) {
+    g_slist_free (GTS_OBJECT (e1)->reserved);
+    GTS_OBJECT (e1)->reserved = NULL;
+    gts_object_destroy (GTS_OBJECT (e1));
+    e1 = NULL;
+  }
+  if (e2->triangles == NULL) {
+    g_slist_free (GTS_OBJECT (e2)->reserved);
+    GTS_OBJECT (e2)->reserved = NULL;
+    gts_object_destroy (GTS_OBJECT (e2));
+    e2 = NULL;
+  }
+  if (e3->triangles == NULL) {
+    g_slist_free (GTS_OBJECT (e3)->reserved);
+    GTS_OBJECT (e3)->reserved = NULL;
+    gts_object_destroy (GTS_OBJECT (e3));
+    e3 = NULL;
+  }
+
+  e56 = gts_edge_new (edge_class, v5, v6);
+  e64 = gts_edge_new (edge_class, v6, v4);
+  e45 = gts_edge_new (edge_class, v4, v5);
+  t->e1 = e56; e56->triangles = g_slist_prepend (e56->triangles, t);
+  t->e2 = e64; e64->triangles = g_slist_prepend (e64->triangles, t);
+  t->e3 = e45; e45->triangles = g_slist_prepend (e45->triangles, t);
+  
+  gts_surface_add_face (s, gts_face_new (s->face_class, e16, e56, e15));
+  gts_surface_add_face (s, gts_face_new (s->face_class, e26, e24, e64));
+  gts_surface_add_face (s, gts_face_new (s->face_class, e45, e34, e35)); 
+}
+
+static void create_array_tessellate (GtsFace * f, GPtrArray * array)
+{
+  g_ptr_array_add (array, f);
+}
+
+/**
+ * gts_surface_tessellate:
+ * @s: a #GtsSurface.
+ * @refine_func: a #GtsRefineFunc.
+ * @refine_data: user data to be passed to @refine_func.
+ *
+ * Tessellate each triangle of @s with 4 triangles:   
+ * the number of triangles is increased by a factor of 4.
+ * http://mathworld.wolfram.com/GeodesicDome.html
+ *
+ * If @refine_func is set to %NULL a mid arc function is used: if
+ * the surface is a polyhedron with the unit sphere as circum sphere,
+ * then gts_surface_tessellate() corresponds to a geodesation step
+ * (see gts_surface_generate_sphere()).
+ * 
+ */
+void gts_surface_tessellate (GtsSurface * s,
+			     GtsRefineFunc refine_func,
+			     gpointer refine_data)
+{
+  GPtrArray * array;
+  guint i;
+
+  g_return_if_fail (s != NULL);
+  
+  if (refine_func == NULL) /* tessellate_surface == geodesate_surface */
+    refine_func = (GtsRefineFunc) unit_sphere_arc_midvertex;
+
+  array = g_ptr_array_new ();
+  gts_surface_foreach_face (s, (GtsFunc) create_array_tessellate, array);
+  for(i = 0; i < array->len; i++)
+    tessellate_face (g_ptr_array_index (array, i),
+		     s, refine_func, refine_data, 
+		     s->vertex_class, s->edge_class);
+  g_ptr_array_free (array, TRUE);
+}
+
+/**
+ * gts_surface_generate_sphere:
+ * @s: a #GtsSurface.
+ * @geodesation_order: a #guint.
+ *
+ * Add a triangulated unit sphere generated by recursive subdivision to @s.
+ * First approximation is an isocahedron; each level of refinement
+ * (@geodesation_order) increases the number of triangles by a factor of 4.
+ * http://mathworld.wolfram.com/GeodesicDome.html
+ *
+ * Returns: @s.
+ */
+GtsSurface * gts_surface_generate_sphere (GtsSurface * s, 
+					  guint geodesation_order)
+{
+  guint cgo; 
+
+  g_return_val_if_fail (s != NULL, NULL);
+  g_return_val_if_fail (geodesation_order != 0, NULL);
+
+  generate_icosahedron (s);
+
+  for (cgo = 1; cgo < geodesation_order; cgo++)
+    gts_surface_tessellate (s, NULL, NULL);
+  
+  return s;
+}
+
 static void foreach_vertex_copy (GtsPoint * p, GtsVertexClass * klass)
 {
   GTS_OBJECT (p)->reserved = gts_vertex_new (klass, p->x, p->y, p->z);
@@ -1868,7 +2233,11 @@ guint gts_surface_face_number (GtsSurface * s)
 {
   g_return_val_if_fail (s != NULL, 0);
 
+#ifdef USE_SURFACE_BTREE
+  return g_tree_nnodes (s->faces);
+#else /* not USE_SURFACE_BTREE */
   return g_hash_table_size (s->faces);
+#endif /* not USE_SURFACE_BTREE */
 }
 
 static void build_list_face (GtsTriangle * t, GSList ** list)
