@@ -28,13 +28,30 @@
 #endif /* HAVE_UNISTD_H */
 #include "gts.h"
 
-typedef enum { NONE, QUALITY, AREA, EPV, INCOMP, FOLD, HEIGHT } ColorOptions;
+#ifndef PI
+# define PI 3.14159265359
+#endif
+
+typedef enum { NONE, 
+	       QUALITY, 
+	       AREA, 
+	       EPV, 
+	       INCOMP, 
+	       FOLD, 
+	       HEIGHT, 
+	       CURVATURE, 
+	       GAUSSIAN } ColorOptions;
 
 ColorOptions color = NONE;
 gboolean autoscale = TRUE;
 gboolean boundary = FALSE;
+gboolean non_manifold = FALSE;
+gint flatten = -1;
+guint height = 0;
+gdouble avg = 0.;
 gdouble min = 0.0, max = 1.0;
 gdouble maxcosine2 = 1.0;
+gboolean verbose = FALSE;
 
 typedef struct {
   gdouble r, g, b;
@@ -125,14 +142,22 @@ static Color colormap_color (Colormap * cmap, gdouble val)
   return c;
 }
 
+static void write_point (GtsPoint * p, FILE * fp)
+{
+  fprintf (fp, "%g %g %g", 
+	   flatten == 0 ? avg : p->x,
+	   flatten == 1 ? avg : p->y,
+	   flatten == 2 ? avg : p->z);
+}
+
 static void foreach_vertex (GtsVertex * v, gpointer * info) 
 {
   FILE * fp = info[0];
   GHashTable * hash = info[1];
   guint * nv = info[2];
+  GtsSurface * s = info[4];
 
-  fprintf (fp, "%g %g %g", 
-	   GTS_POINT (v)->x, GTS_POINT (v)->y, GTS_POINT (v)->z);
+  write_point (GTS_POINT (v), fp);
   switch (color) {
   case EPV: {
     Color c = colormap_color (colormap,
@@ -142,34 +167,82 @@ static void foreach_vertex (GtsVertex * v, gpointer * info)
   }
   case HEIGHT: {
     Color c = colormap_color (colormap,
-			      (GTS_POINT (v)->z - min)/(max - min));
+	       ((&GTS_POINT (v)->x)[height] - avg - min)/(max - min));
     fprintf (fp, " %g %g %g 1.0\n", c.r, c.g, c.b);
     break;
   }
+  case CURVATURE: {
+    gdouble c = 0.;
+    Color color;
+    GtsVector n;
+
+    if (!gts_vertex_is_boundary (v, s) &&
+	gts_vertex_mean_curvature_normal (v, s, n)) 
+      c = gts_vector_norm (n)/2.;
+    color = colormap_color (colormap, (c - min)/(max - min));
+    fprintf (fp, " %g %g %g 1.0\n", color.r, color.g, color.b);
+    break;
+  }
+  case GAUSSIAN: {
+    gdouble gc = 0.;
+    Color color;
+
+    if (!gts_vertex_is_boundary (v, s))
+      gts_vertex_gaussian_curvature (v, s, &gc);
+    color = colormap_color (colormap, (gc - min)/(max - min));
+    fprintf (fp, " %g %g %g 1.0\n", color.r, color.g, color.b);
+    break;
+  }
   default:
     fprintf (fp, "\n");
   }
   g_hash_table_insert (hash, v, GUINT_TO_POINTER (++(*nv)));
 }
 
-static void foreach_edge (GtsEdge * e, gpointer info)
+static void foreach_boundary_edge (GtsEdge * e, gpointer info)
 {
   FILE * fp = info;
 
-  if (gts_edge_is_boundary (e, NULL))
-    fprintf (fp, "VECT 1 2 0 2 0 %g %g %g %g %g %g\n",
-	     GTS_POINT (GTS_SEGMENT (e)->v1)->x, 
-	     GTS_POINT (GTS_SEGMENT (e)->v1)->y, 
-	     GTS_POINT (GTS_SEGMENT (e)->v1)->z,
-	     GTS_POINT (GTS_SEGMENT (e)->v2)->x, 
-	     GTS_POINT (GTS_SEGMENT (e)->v2)->y, 
-	     GTS_POINT (GTS_SEGMENT (e)->v2)->z);
+  if (gts_edge_is_boundary (e, NULL)) {
+    fputs ("VECT 1 2 0 2 0 ", fp);
+    write_point (GTS_POINT (GTS_SEGMENT (e)->v1), fp);
+    fputc (' ', fp);
+    write_point (GTS_POINT (GTS_SEGMENT (e)->v2), fp);
+    fputc ('\n', fp);
+  }
+}
+
+static void foreach_feature_edge (GtsEdge * e, gdouble * angle)
+{
+  if (g_slist_length (e->triangles) == 2 &&
+      fabs (gts_triangles_angle (e->triangles->data, e->triangles->next->data))
+      < *angle) {
+    fputs ("VECT 1 2 0 2 0 ", stdout);
+    write_point (GTS_POINT (GTS_SEGMENT (e)->v1), stdout);
+    fputc (' ', stdout);
+    write_point (GTS_POINT (GTS_SEGMENT (e)->v2), stdout);
+    fputc ('\n', stdout);
+  }
+}
+
+static void foreach_non_manifold_edge (GtsEdge * e, gpointer info)
+{
+  FILE * fp = info;
+
+  if (g_slist_length (e->triangles) > 2) {
+    fputs ("VECT 1 2 0 2 0 ", fp);
+    write_point (GTS_POINT (GTS_SEGMENT (e)->v1), fp);
+    fputc (' ', fp);
+    write_point (GTS_POINT (GTS_SEGMENT (e)->v2), fp);
+    fputc ('\n', fp);
+  }
 }
 
 static void foreach_triangle (GtsTriangle * t, gpointer * info)
 {
   FILE * fp = info[0];
   GHashTable * hash = info[1];
+  guint * nfold = info[3];
   guint p1 = 0, p2 = 0, p3 = 0;
 
   p1 = GPOINTER_TO_UINT (g_hash_table_lookup (hash, GTS_SEGMENT (t->e1)->v1));
@@ -260,8 +333,10 @@ static void foreach_triangle (GtsTriangle * t, gpointer * info)
 				       GTS_SEGMENT (e3)->v1,
 				       GTS_SEGMENT (e3)->v2,
 				       maxcosine2);      
-    if (fold)
+    if (fold) {
+      (*nfold)++;
       fputs (" 1. 0. 0.", fp);
+    }
     fputc ('\n', fp);
   }
   default:
@@ -269,12 +344,40 @@ static void foreach_triangle (GtsTriangle * t, gpointer * info)
   }
 }
 
+static void min_max_curvature (GtsVertex * v, gpointer * data)
+{
+  gdouble * min = data[0];
+  gdouble * max = data[1];
+  GtsSurface * s = data[2];
+  GtsVector n;
+  gdouble c = 0.;
+
+  if (!gts_vertex_is_boundary (v, s) && 
+      gts_vertex_mean_curvature_normal (v, s, n)) 
+    c = gts_vector_norm (n)/2.;
+  if (c < *min) *min = c;
+  if (c > *max) *max = c;
+}
+
+static void min_max_gaussian (GtsVertex * v, gpointer * data)
+{
+  gdouble * min = data[0];
+  gdouble * max = data[1];
+  GtsSurface * s = data[2];
+  gdouble gc = 0.;
+
+  if (!gts_vertex_is_boundary (v, s))
+    gts_vertex_gaussian_curvature (v, s, &gc);
+  if (gc < *min) *min = gc;
+  if (gc > *max) *max = gc;
+}
+
 static void oogl_surface (GtsSurface * s, FILE * fptr)
 {
-  gpointer info[3];
+  gpointer info[5];
   GtsSurfaceStats stats;
   GtsSurfaceQualityStats qstats;
-  guint np = 0;
+  guint np = 0, nfold = 0;
 
   g_return_if_fail (s != NULL);
   g_return_if_fail (fptr != NULL);
@@ -282,6 +385,8 @@ static void oogl_surface (GtsSurface * s, FILE * fptr)
   info[0] = fptr;
   info[1] = g_hash_table_new (NULL, NULL);
   info[2] = &np;
+  info[3] = &nfold;
+  info[4] = s;
 
   gts_surface_stats (s, &stats);
   switch (color) {
@@ -308,16 +413,45 @@ static void oogl_surface (GtsSurface * s, FILE * fptr)
   case HEIGHT:
     if (autoscale) {
       GtsBBox * bbox = gts_bbox_surface (gts_bbox_class (), s);
-      min = bbox->z1;
-      max = bbox->z2;
+
+      min = (&bbox->x1)[height] - avg;
+      max = (&bbox->x2)[height] - avg;
       gts_object_destroy (GTS_OBJECT (bbox));
     }
     break;
+  case CURVATURE:
+    if (autoscale) {
+      gpointer data[3];
+
+      min = G_MAXDOUBLE;
+      max = - G_MAXDOUBLE;
+      data[0] = &min;
+      data[1] = &max;
+      data[2] = s;
+      gts_surface_foreach_vertex (s, (GtsFunc) min_max_curvature, data);
+    }
+    break;
+  case GAUSSIAN:
+    if (autoscale) {
+      gpointer data[3];
+
+      min = G_MAXDOUBLE;
+      max = - G_MAXDOUBLE;
+      data[0] = &min;
+      data[1] = &max;
+      data[2] = s;
+      gts_surface_foreach_vertex (s, (GtsFunc) min_max_gaussian, data);
+    }
+    break;
   default:
+    ;
   }
+  if (verbose)
+    fprintf (stderr, "scalar min: %g max: %g\n", min, max);
   if (max == min)
     max = min + 1.;
-  if (color == EPV || color == HEIGHT)
+  if (color == EPV || color == HEIGHT || 
+      color == CURVATURE || color == GAUSSIAN)
     fputs ("COFF ", fptr);
   else
     fputs ("OFF ", fptr);
@@ -325,8 +459,10 @@ static void oogl_surface (GtsSurface * s, FILE * fptr)
 	   stats.edges_per_vertex.n, 
 	   stats.n_faces,
 	   stats.faces_per_edge.n);
-  gts_surface_foreach_vertex (s, (GtsFunc)foreach_vertex, info);
-  gts_surface_foreach_face (s, (GtsFunc)foreach_triangle, info);
+  gts_surface_foreach_vertex (s, (GtsFunc) foreach_vertex, info);
+  gts_surface_foreach_face (s, (GtsFunc) foreach_triangle, info);
+  if (verbose && color == FOLD)
+    fprintf (stderr, "# Number of folds: %u\n", nfold);
 
   g_hash_table_destroy (info[1]);
 }
@@ -408,7 +544,7 @@ static void write_isolines (GtsSurface * surface,
 			       gts_edge_new (gts_edge_class (), v3, v1)));
   ptree = gts_bb_tree_surface (plane);
   si = gts_surface_inter_new (gts_surface_inter_class (),
-			      surface, plane, stree, ptree);
+			      surface, plane, stree, ptree, FALSE, FALSE);
   chains = chain_segments (si->edges);
 
   i = chains;
@@ -440,7 +576,8 @@ static void write_isolines (GtsSurface * surface,
     while (j) {
       GtsPoint * p = j->data;
 
-      fprintf (fp, "%g %g %g\n", p->x, p->y, p->z);
+      write_point (p, fp);
+      fputc ('\n', fp);
       j = j->next;
     }
     g_slist_free (chain);
@@ -452,26 +589,36 @@ static void write_isolines (GtsSurface * surface,
   gts_object_destroy (GTS_OBJECT (plane));
 }
 
+static void sum_flatten (GtsPoint * p)
+{
+  avg += (&p->x)[flatten];
+}
+
 int main (int argc, char * argv[])
 {
   GtsSurface * surface;
-  gboolean verbose = FALSE;
   gboolean nosurface = FALSE;
   int c = 0;
-  guint line;
   guint isolines = 0;
+  GtsFile * fp;
+  gdouble feature = 0.;
 
   colormap = colormap_jet (); /* default */
   while (c != EOF) {
 #ifdef HAVE_GETOPT_LONG
     static struct option long_options[] = {
+      {"gaussian", no_argument, NULL, 'g'},
+      {"curvature", no_argument, NULL, 'C'},
+      {"feature", required_argument, NULL, 'e'},
+      {"flatten", required_argument, NULL, 'F'},
       {"quality", no_argument, (int *)&color, QUALITY},
       {"area", no_argument, (int *)&color, AREA},
       {"incomp", no_argument, (int *)&color, INCOMP},
       {"fold", required_argument, NULL, 'f'},
       {"epv", no_argument, (int *)&color, EPV},
-      {"height", no_argument, (int *)&color, HEIGHT},
+      {"height", required_argument, NULL, 'H'},
       {"boundary", no_argument, &boundary, TRUE},
+      {"non-manifold", no_argument, &non_manifold, TRUE},
       {"isolines", required_argument, NULL, 'i'},
       {"cmap", required_argument, NULL, 'c'},
       {"help", no_argument, NULL, 'h'},
@@ -482,11 +629,31 @@ int main (int argc, char * argv[])
       {"verbose", no_argument, NULL, 'v'}
     };
     int option_index = 0;
-    switch ((c = getopt_long (argc, argv, "c:hm:M:rvf:i:n",
+    switch ((c = getopt_long (argc, argv, "c:hm:M:rvf:i:nF:e:H:Cg",
 			      long_options, &option_index))) {
 #else /* not HAVE_GETOPT_LONG */
-    switch ((c = getopt (argc, argv, "c:hm:M:rvf:i:n"))) {
+    switch ((c = getopt (argc, argv, "c:hm:M:rvf:i:nF:e:H:Cg"))) {
 #endif /* not HAVE_GETOPT_LONG */
+    case 'g': /* gaussian */
+      color = GAUSSIAN;
+      break;
+    case 'C': /* curvature */
+      color = CURVATURE;
+      break;
+    case 'H': /* height */
+      color = HEIGHT;
+      height = atoi (optarg);
+      if (height > 2)
+	height = 2;
+      break;
+    case 'e': /* feature */
+      feature = atof (optarg)*PI/180.;
+      break;
+    case 'F': /* flatten */
+      flatten = atoi (optarg);
+      if (flatten < 0.) flatten = 0;
+      else if (flatten > 2) flatten = 2;
+      break;
     case 'c': { /* colormap file */
       FILE * fptr = fopen (optarg, "rt");
       if (!fptr) {
@@ -503,7 +670,7 @@ int main (int argc, char * argv[])
       break;
     case 'f': /* fold */
       color = FOLD;
-      maxcosine2 = cos (atof (optarg)*3.14159265359/180.);
+      maxcosine2 = cos (atof (optarg)*PI/180.);
       maxcosine2 *= maxcosine2;
       break;
     case 'm': /* minimum value */
@@ -532,14 +699,19 @@ int main (int argc, char * argv[])
 	     "  -f VAL  --fold=VAL     color faces which make an angle smaller\n" 
 	     "                         than VAL degrees with any of their neighbors\n"
 	     "          --epv          color vertices according to number of edges per vertex\n"
-	     "          --height       color vertices according to their height\n"
+	     "  -H C    --height=C     color vertices according to their C coordinate\n"
+	     "  -g      --gaussian     color vertices according to Gaussian curvature\n"
+	     "  -C      --curvature    color vertices according to mean curvature\n"
 	     "          --boundary     output boundary edges\n"
+             "  -e A    --feature=A    output `feature' edges defined by angle A\n"
+             "          --non-manifold output non-manifold edges\n"
 	     "  -i N    --isolines=N   draw N isolines (levels of constant altitude)\n"
 	     "          --cmap=FILE    load FILE as colormap\n"
 	     "  -m VAL, --min=VAL      use VAL as minimum scaling value\n"
 	     "  -M VAL, --max=VAL      use VAL as maximum scaling value\n"
 	     "  -r,     --reverse      reverse colormap\n"
              "  -n      --nosurface    do not output surface\n"
+	     "  -F C    --flatten=C    set C coordinate to average value\n"
 	     "  -v,     --verbose      display surface statistics\n"
 	     "  -h,     --help         display this help and exit\n"
 	     "\n"
@@ -560,31 +732,57 @@ int main (int argc, char * argv[])
 			     gts_face_class (),
 			     gts_edge_class (),
 			     gts_vertex_class ());
-  line = gts_surface_read (surface, stdin);
-
-  if (line) {
+  fp = gts_file_new (stdin);
+  if (gts_surface_read (surface, fp)) {
     fputs ("gts2oogl: the file on standard input is not a valid GTS file\n", 
 	   stderr);
-    fprintf (stderr, "Format error at line %d\n", line);
+    fprintf (stderr, "stdin:%d:%d: %s\n", fp->line, fp->pos, fp->error);
     return 1;
   }
   
   if (verbose)
     gts_surface_print_stats (surface, stderr);
 
-  if (boundary || isolines) {
+  if (flatten >= 0) {
+    gts_surface_foreach_vertex (surface, (GtsFunc) sum_flatten, NULL);
+    if (avg != 0.)
+      avg /= gts_surface_vertex_number (surface);
+  }
+
+  if (boundary || non_manifold || isolines || feature > 0.) {
     if (!nosurface) {
       puts ("(geometry \"surface\" { = ");
       oogl_surface (surface, stdout);
       puts ("})");
     }
     if (boundary) {
-      if (!nosurface || isolines)
+      if (!nosurface || isolines || non_manifold)
 	puts ("(geometry \"boundary\" { = ");
       puts ("LIST {");
-      gts_surface_foreach_edge (surface, (GtsFunc) foreach_edge, stdout);
+      gts_surface_foreach_edge (surface, (GtsFunc) foreach_boundary_edge, 
+				stdout);
       puts ("}");
-      if (!nosurface || isolines)
+      if (!nosurface || isolines || non_manifold)
+	puts ("})");
+    }
+    if (feature > 0.) {
+      if (!nosurface || isolines || non_manifold)
+	puts ("(geometry \"feature\" { = ");
+      puts ("LIST {");
+      gts_surface_foreach_edge (surface, (GtsFunc) foreach_feature_edge, 
+				&feature);
+      puts ("}");
+      if (!nosurface || isolines || non_manifold)
+	puts ("})");
+    }
+    if (non_manifold) {
+      if (!nosurface || isolines || boundary)
+	puts ("(geometry \"non-manifold\" { = ");
+      puts ("LIST {");
+      gts_surface_foreach_edge (surface, (GtsFunc) foreach_non_manifold_edge, 
+				stdout);
+      puts ("}");
+      if (!nosurface || isolines || boundary)
 	puts ("})");
     }
     if (isolines > 0) {
